# Copyright (C) 2024 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Engineering units and converters.
"""

from __future__ import annotations

from numbers import Number

import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline

from .logging_ import logger
from .plt_ import plt, setup_gca


class ConversionBase:

    """Abstract base class for a generic conversion.
    """

    # pylint: disable=too-few-public-methods

    def _conversion_function(self, raw) -> None:
        """Conversion function, to be reimplemented in derived classes.
        """
        raise NotImplementedError

    def __call__(self, raw):
        """Special dunder method for executing the actual conversion.
        """
        return self._conversion_function(raw)


class LinearConversion(ConversionBase):

    """Linear conversion.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, slope: float, intercept: float = 0) -> None:
        """Constructor.
        """
        self.slope = slope
        self.intercept = intercept

    def _conversion_function(self, raw):
        """Overloaded method.
        """
        return self.slope * raw + self.intercept


# class PolynomialConversion(ConversionBase):
#
#     """Polynomial conversion.
#     """
#
#     pass


class SplineConversion(ConversionBase):

    """Spline conversion.

    This is performing a simple spline interpolation of the physical values vs. the
    raw values.

    Arguments
    ---------
    raw : array_like
        The array of raw values.

    physical : array_like
        The array of physical values

    k : int (default 3)
        The degree of the interpolating spline.
    """

    def __init__(self, raw: np.array, physical: np.array, k: int = 3) -> None:
        """Constructor.
        """
        self._raw, self._physical = self._process_input(raw, physical)
        self._spline = InterpolatedUnivariateSpline(self._raw, self._physical, k=k)

    @staticmethod
    def _process_input(raw, physical) -> tuple[np.array, np.array]:
        """Sort and remove duplicates from the input arrays.

        In order for us to be able to build the spline, the x (raw) values must
        be passed in ascending order, and the y (physical) values must be (potentially)
        shuffled accordingly.

        Looking around on the internet, it seems like simply checking if an array
        is sorted is almost as expensive as calling argsort or unique, see, e.g.,
        https://github.com/numpy/numpy/issues/8392
        and therefore we just go ahead and re-sort everything and remove duplicates
        even when there is nothing to do.
        """
        raw, _index = np.unique(raw, return_index=True)
        physical = physical[_index]
        return raw, physical

    @staticmethod
    def read_data(file_path: str, col_raw: int = 0, col_physical: int = 1, **kwargs):
        """Read data from file.

        Arguments
        ---------
        file_path : str
            The path to the input text file containing the data.

        col_raw : int
            The index of the column containing the raw values.

        col_physical : int
            The index of the column containing the physical values.

        kwrargs : dict
            optional keyword arguments passed to ``np.loadtxt()``
        """
        logger.info(f"Reading conversion data from {file_path}...")
        raw, physical = np.loadtxt(file_path, usecols=(col_raw, col_physical),
                                   unpack=True, **kwargs)
        logger.info(f"Done, {len(raw)} data point(s) found.")
        return raw, physical

    @classmethod
    def from_file(cls, file_path: str, col_raw: int = 0, col_physical: int = 1,
                  k: int = 3) -> None:
        """Read the data points for the spline from a file.
        """
        data = cls.read_data(file_path, col_raw, col_physical)
        return cls(*data, k)

    def plot(self):
        """Plot the interpolating spline.
        """
        plt.plot(self._raw, self._physical)
        setup_gca(xlabel="Raw units", ylabel="Physical units")

    def _conversion_function(self, raw):
        """Overloaded method.
        """
        # If we are passing a single number, we cast the value to float in order to
        # avoid getting a 0-dim array as output from the spline evaluation, which
        # would potentially be a nuisance to handle downstream.
        if isinstance(raw, Number):
            return float(self._spline(raw))
        return self._spline(raw)


class ThermistorConversion(SplineConversion):

    """Specific conversion for a thermistor.
    """

    def __init__(self, temperature: np.array, resistance: np.array, shunt_resistance: float,
                 adc_num_bits: int, k: int = 3) -> None:
        """Constructor.
        """
        # pylint: disable=too-many-arguments
        adc = (2**adc_num_bits - 1) * shunt_resistance / (resistance + shunt_resistance)
        super().__init__(adc, temperature, k)

    @classmethod
    def from_file(cls, file_path: str, shunt_resistance: float, adc_num_bits: int,
                  col_temperature: int = 0, col_resistance: int = 1, k: int = 3) -> None:
        """Read the data points for the spline from file.
        """
        # pylint: disable=arguments-renamed, too-many-arguments
        data = cls.read_data(file_path, col_temperature, col_resistance)
        return cls(*data, shunt_resistance, adc_num_bits, k)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import importlib.metadata

from baldaquin import __version__, __name__ as __package_name__


# Get package metadata.
_metadata = importlib.metadata.metadata(__package_name__)


# --- Project information ---
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = __package_name__
author = _metadata["Author-email"]
copyright = f"2025-%Y, {author}"
version = __version__
release = version

# --- General configuration ---
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.imgmath",
    "sphinx.ext.extlinks",
    "sphinxcontrib.programoutput",
]
autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
    "undoc-members": True,
    "private-members": True
}
todo_include_todos = True

# Options for syntax highlighting.
pygments_style = "default"
pygments_dark_style = "default"

# Options for internationalization.
language = "en"

# Options for markup.
rst_prolog = f"""
"""

# Options for source files.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Options for templating.
templates_path = ["_templates"]


# --- Options for HTML output ---
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinxawesome_theme"
html_theme_options = {
    "awesome_external_links": True,
}
#html_logo = "_static/baldaquin_logo_small_light.png"
#html_favicon = "_static/favicon.ico"
html_permalinks_icon = "<span>#</span>"
html_static_path = ["_static"]




# Copyright (C) 2022 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for runctrl.py
"""

import time

import pytest

from baldaquin.env import BALDAQUIN_SCRATCH
from baldaquin.pkt import PacketStatistics
from baldaquin.runctrl import FiniteStateMachineBase, InvalidFsmTransitionError, RunReport
from baldaquin.timeline import Timeline


def test_finite_state_machine():
    """This creates a mock FSM where all the relevant hooks are no-op, and tests
    for the proper states after transitions, as well as the exceptions signaling
    the invalid ones.
    """
    # Create a virtual FSM.
    fsm = FiniteStateMachineBase()
    # Override all the necessary hooks to that we can actually have transitions.
    fsm.setup = lambda: True
    fsm.teardown = lambda: True
    fsm.start_run = lambda: True
    fsm.stop_run = lambda: True
    fsm.pause = lambda: True
    fsm.stop = lambda: True
    fsm.resume = lambda: True
    # Start in the RESET state.
    assert fsm.is_reset()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_running()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_reset()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_paused()
    # Transition to STOPPED.
    fsm.set_stopped()
    assert fsm.is_stopped()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_stopped()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_paused()
    # Transition to RUNNING.
    fsm.set_running()
    assert fsm.is_running()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_reset()
    # Transition to PAUSED.
    fsm.set_paused()
    assert fsm.is_paused()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_reset()
    # And now back to back and forth from/to STOPPED and to the RESET state...
    fsm.set_running()
    fsm.set_stopped()
    fsm.set_running()
    fsm.set_paused()
    fsm.set_stopped()
    fsm.set_reset()
    assert fsm.is_reset()


def test_report():
    """Small unit test for the run report.
    """
    # Create a report.
    timeline = Timeline()
    start_timestamp = timeline.latch()
    time.sleep(0.5)
    stop_timestamp = timeline.latch()
    stats = PacketStatistics(10, 10, 100)
    report = RunReport("0.3.1", 101, 66, start_timestamp, stop_timestamp,
                       "Test project", "TestApplication", stats)

    # Make sure that the serialization/deserialization roundtrips.
    kwargs = report.to_dict()
    twin = RunReport.from_dict(**kwargs)
    assert twin == report
    assert twin.baldaquin_version == report.baldaquin_version
    assert twin.test_stand_id == report.test_stand_id
    assert twin.run_id == report.run_id
    assert twin.start_timestamp == report.start_timestamp
    assert twin.stop_timestamp == report.stop_timestamp
    assert twin.application_name == report.application_name
    assert twin.statistics == report.statistics

    # Test the file IO.
    file_path = BALDAQUIN_SCRATCH / "test_report.json"
    report.save(file_path)
    twin = RunReport.load(file_path)
    assert twin == report
    assert twin.baldaquin_version == report.baldaquin_version
    assert twin.test_stand_id == report.test_stand_id
    assert twin.run_id == report.run_id
    assert twin.start_timestamp == report.start_timestamp
    assert twin.stop_timestamp == report.stop_timestamp
    assert twin.application_name == report.application_name
    assert twin.statistics == report.statistics



# Copyright (C) 2022 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for config.py
"""

import json

import pytest

from baldaquin import config
from baldaquin.env import BALDAQUIN_DATA, BALDAQUIN_ENCODING


def _test_base_match(type_name, value, **constraints):
    """Base test function where we expect the parameter to match the input value.
    """
    p = config.ConfigurationParameter("parameter", type_name, value, "", **constraints)
    assert p.value == value


def _test_base_mismatch(type_name, value, **constraints):
    """Base test function where we expect the parameter to match the input value.
    """
    with pytest.raises(RuntimeError):
        config.ConfigurationParameter("parameter", type_name, value, "", **constraints)


def test_parameter_bool():
    """Test possible setting for int parameters.
    """
    for value in (True, False):
        _test_base_match(bool, value)
    for value in (1, 1., "test"):
        _test_base_mismatch(bool, value)


def test_parameter_int():
    """Test possible settings for int parameters.
    """
    _test_base_match(int, 100)
    _test_base_match(int, 100, min=0, max=1000)
    _test_base_mismatch(int, 100, min=200)
    _test_base_mismatch(int, 100, max=0)
    _test_base_match(int, 100, choices=(99, 100, 101))
    _test_base_mismatch(int, 100, choices=(99, 101))
    _test_base_match(int, 100, step=10)
    _test_base_mismatch(int, 100, step=13)
    _test_base_match(int, 100, min=87, step=13)


def test_parameter_float():
    """Test possible settings for int parameters.
    """
    _test_base_match(float, 1.)
    _test_base_match(float, 1., min=0., max=10.)
    _test_base_mismatch(float, 1., min=10.)
    _test_base_mismatch(float, 1., max=0.)


def test_parameter_str():
    """Test possible settings for int parameters.
    """
    _test_base_match(str, "howdy?")
    _test_base_match(str, "eggs", choices=("eggs", "cheese"))
    _test_base_mismatch(str, "ham", choices=("eggs", "cheese"))
    for value in (True, 1, 1.):
        _test_base_mismatch(str, value)


def test_configuration_sections():
    """Test the configuration sections.
    """
    section = config.LoggingConfigurationSection()
    section.set_value("terminal_level", "DEBUG")
    assert section.value("terminal_level") == "DEBUG"
    with pytest.raises(RuntimeError):
        section.set_value("terminal_level", "PHONY")
    assert section.value("terminal_level") == "DEBUG"
    section.set_value("terminal_level", "INFO")
    assert section.value("terminal_level") == "INFO"
    with pytest.raises(RuntimeError):
        section.set_value("dummy", "PHONY")
    _ = section.as_dict()
    section = config.BufferingConfigurationSection()
    section = config.MulticastConfigurationSection()


def test_application_configuration():
    """Test on the basic baldaquin configuration.
    """
    conf = config.UserApplicationConfiguration()
    file_path = BALDAQUIN_DATA / "baldaquin.cfg"
    conf.save(file_path)
    conf["Logging"].set_value("terminal_level", "CRITICAL")
    assert conf["Logging"].value("terminal_level") == "CRITICAL"
    conf.update_from_file(file_path)
    assert conf["Logging"].value("terminal_level") == "DEBUG"


def _write_configuration_dict(file_path: str, data: dict) -> None:
    """Utility function to write a configuration dictionary to disk.
    """
    with open(file_path, "w", encoding=BALDAQUIN_ENCODING) as output_file:
        output_file.write(json.dumps(data, indent=4))


def test_resilience():
    """Test the resilience of the configuration mechanism to format changes.
    """
    conf = config.UserApplicationConfiguration()
    file_path = BALDAQUIN_DATA / "resilience.cfg"
    # Add an unknown section.
    data = conf.as_dict()
    data["Fake"] = {"fake_field": 3}
    _write_configuration_dict(file_path, data)
    conf.update_from_file(file_path)
    # Add an unknown field in a legitimate section.
    data = conf.as_dict()
    data["Logging"] = {"fake_field": 3}
    _write_configuration_dict(file_path, data)
    conf.update_from_file(file_path)
    # Set a wrong value for a legitimate parameter.
    data = conf.as_dict()
    data["Logging"] = {"terminal_level": "FAKE"}
    _write_configuration_dict(file_path, data)
    conf.update_from_file(file_path)



# Copyright (C) 2024 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Plasduino shield information.

See the old plasduino repo for all the gory details:
https://bitbucket.org/lbaldini/plasduino/src/main/shields/
"""


class Lab1:

    """Hardware description for the Lab1 shield, version 2.
    """

    # pylint: disable=too-few-public-methods

    VERSION = 2

    # Analog inputs for the temperature measurements.
    TEMPMON_PIN_1 = 0
    TEMPMON_PIN_2 = 1
    TEMPMON_PINS = (TEMPMON_PIN_1, TEMPMON_PIN_2)
    SHUNT_RESISTANCE = 10.  # kOhm

    # Analog inputs for the pendulumview application.
    PENDVIEW_PIN1 = 4
    PENDVIEW_PIN2 = 5
    PENDVIEW_PINS = (PENDVIEW_PIN1, PENDVIEW_PIN2)

    # Digital inputs for the time measurements.
    TIMER_PIN_1 = 0
    TIMER_PIN_2 = 1
    TIMER_PINS = (TIMER_PIN_1, TIMER_PIN_2)



# Copyright (C) 2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Silly application with strip charts.
"""

from aptapy.plotting import VerticalCursor
from aptapy.strip import StripChart

from baldaquin import silly
from baldaquin.__qt__ import QtWidgets
from baldaquin.gui import bootstrap_window
from baldaquin.pkt import AbstractPacket
from baldaquin.runctrl import RunControlBase
from baldaquin.silly.common import (
    SillyConfiguration,
    SillyMainWindow,
    SillyPacket,
    SillyRunControl,
    SillyUserApplicationBase,
)


class MainWindow(SillyMainWindow):

    def __init__(self, parent: QtWidgets.QWidget = None) -> None:
        """Constructor.
        """
        super().__init__()
        self.strip_tab = self.add_plot_canvas_tab("Strip charts")
        self.tab_widget.setCurrentWidget(self.strip_tab)

    def setup_user_application(self, user_application):
        """Overloaded method.
        """
        super().setup_user_application(user_application)
        # This line is ugly, and we should find a better way to provide the user
        # application with access to the axes objects in the plotting widgets.
        user_application.axes = self.strip_tab.axes
        self.strip_tab.register(user_application.strip_chart)


class SillyStrip(SillyUserApplicationBase):

    """Simple user application for testing purposes.
    """

    NAME = "Silly strip chart display"
    CONFIGURATION_CLASS = SillyConfiguration
    CONFIGURATION_FILE_PATH = silly.SILLY_APP_CONFIG / "silly_strip.cfg"

    def __init__(self):
        """Overloaded constructor.
        """
        super().__init__()
        self.strip_chart = StripChart(max_length=100, label="Random data",
                                      xlabel="Trigger ID", ylabel="PHA")
        self.axes = None
        self._cursor = None

    def configure(self) -> None:
        """Overloaded method.
        """
        self.strip_chart.clear()

    def process_packet(self, packet_data: bytes) -> AbstractPacket:
        """Dumb data processing routine---print out the actual event.
        """
        packet = SillyPacket.unpack(packet_data)
        self.strip_chart.put(packet.trigger_id, packet.pha)
        return packet

    def pre_start(self, run_control: RunControlBase) -> None:
        """Overloaded method.
        """
        if self._cursor is not None:
            self._cursor.deactivate()
            self._cursor = None

    def post_stop(self, run_control: RunControlBase) -> None:
        """Overloaded method.

        Note the underlying strip charts are unlimited, and there is no need to re-read
        data from disk. We just need to re-draw the strip chart and enable the cursor.

        Also, it would be nice to understand exactly why we need to clear the axis and
        re-plot the strip chart for the cursor to pick the right color. In my mind
        the last_line_color() should do this behind the scenes, but clearly it does not.
        """
        self.axes.clear()
        self._cursor = VerticalCursor(self.axes)
        self.strip_chart.plot(self.axes)
        self._cursor.add_marker(self.strip_chart.spline())
        self.axes.figure.canvas.draw()
        self._cursor.activate()


def main() -> None:
    """Main entry point.
    """
    bootstrap_window(MainWindow, SillyRunControl(), SillyStrip())


if __name__ == "__main__":
    main()



# Copyright (C) 2024--25 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Plasduino pendulum application.
"""

from __future__ import annotations

import inspect
from dataclasses import dataclass
from pathlib import Path

from baldaquin import plasduino
from baldaquin.buf import WriteMode
from baldaquin.env import BALDAQUIN_ENCODING
from baldaquin.gui import MainWindow, SimpleControlBar, bootstrap_window
from baldaquin.logging_ import logger
from baldaquin.pkt import AbstractPacket, Edge, PacketFile
from baldaquin.plasduino import PLASDUINO_APP_CONFIG
from baldaquin.plasduino.common import (
    PlasduinoDigitalConfiguration,
    PlasduinoDigitalEventHandler,
    PlasduinoDigitalUserApplicationBase,
    PlasduinoRunControl,
)
from baldaquin.plasduino.protocol import COMMENT_PREFIX, TEXT_SEPARATOR, DigitalTransition
from baldaquin.runctrl import RunControlBase


class AppMainWindow(MainWindow):

    """Application graphical user interface.
    """

    _PROJECT_NAME = plasduino.PROJECT_NAME
    _CONTROL_BAR_CLASS = SimpleControlBar


@dataclass
class Oscillation:

    """Small convenience class to represent post-processed data.
    """

    average_time: float
    period: float
    transit_time: float

    @staticmethod
    def text_header(prefix: str = COMMENT_PREFIX, creator: str = None) -> str:
        """Text header for the post-processed output file.
        """
        return f"{AbstractPacket.text_header(prefix, creator)}" \
               f"{prefix}Time [s], Period [s], Transit time [s]\n"

    def to_text(self, separator: str = TEXT_SEPARATOR) -> str:
        """Text representation for the output file.
        """
        return f"{self.average_time:.6f}{separator}{self.period:.6f}" \
               f"{separator}{self.transit_time:.6f}\n"


class Pendulum(PlasduinoDigitalUserApplicationBase):

    """Simplest possible user application for testing purposes.
    """

    NAME = "Pendulum"
    CONFIGURATION_CLASS = PlasduinoDigitalConfiguration
    CONFIGURATION_FILE_PATH = PLASDUINO_APP_CONFIG / "plasduino_pendulum.cfg"
    EVENT_HANDLER_CLASS = PlasduinoDigitalEventHandler

    def pre_start(self, run_control: RunControlBase) -> None:
        """Overloaded method.

        Here we are simply adding a text sink to the underlying data buffer to
        write the packets in text form.
        """
        file_path = Path(f"{run_control.output_file_path_base()}_data.txt")
        self.event_handler.add_custom_sink(file_path, WriteMode.TEXT, DigitalTransition.to_text,
                                           DigitalTransition.text_header(creator=self.NAME))

    @staticmethod
    def _secs_avg(data: tuple[DigitalTransition], i: int, j: int) -> float:
        """Convenience function for calculating the average value of the `seconds`
        fields for two given indices in a tuple of digital transitions.

        This is used in the post-processing functions, and it is meant to reduce
        boilerplate code and improve readability.
        """
        return 0.5 * (data[i].seconds + data[j].seconds)

    @staticmethod
    def _secs_diff(data: tuple[DigitalTransition], i: int, j: int) -> float:
        """Convenience function for calculating the difference of the `seconds`
        fields for two given indices in a tuple of digital transitions.

        This is used in the post-processing functions, and it is meant to reduce
        boilerplate code and improve readability.
        """
        return data[i].seconds - data[j].seconds

    @staticmethod
    def _postprocess_data_simple(data: tuple[DigitalTransition]) -> list[Oscillation]:
        """Simple data postprocessing.

        We do not attempt any average for the transit time and, therefore, expect
        to observe small "jumps" up and down if the optical gate is not aligned with the
        pendulum.
        """
        logger.info(f"Running Pendulum.{inspect.currentframe().f_code.co_name}()...")
        oscillations = []
        for i in range(3, len(data) - 1, 2):
            average_time = Pendulum._secs_avg(data, i, i - 1)
            transit_time = Pendulum._secs_diff(data, i, i - 1)
            period = Pendulum._secs_diff(data, i + 1, i - 3)
            oscillations.append(Oscillation(average_time, period, transit_time))
        return oscillations

    @staticmethod
    def _postprocess_data_smooth(data: tuple[DigitalTransition]) -> list[Oscillation]:
        """Slightly more advanced data postprocessing, with a little bit of averaging.
        """
        logger.info(f"Running Pendulum.{inspect.currentframe().f_code.co_name}()...")
        oscillations = []
        for i in range(5, len(data) - 3, 2):
            t1 = Pendulum._secs_avg(data, i - 4, i - 5)
            t2 = Pendulum._secs_avg(data, i - 2, i - 3)
            t3 = Pendulum._secs_avg(data, i, i - 1)
            t4 = Pendulum._secs_avg(data, i + 2, i + 1)
            dt2 = Pendulum._secs_diff(data, i - 2, i - 3)
            dt3 = Pendulum._secs_diff(data, i, i - 1)
            average_time = 0.5 * (t2 + t3)
            transit_time = 0.5 * (dt2 + dt3)
            period = 0.5 * (t3 - t1 + t4 - t2)
            oscillations.append(Oscillation(average_time, period, transit_time))
        return oscillations

    @staticmethod
    def postprocess_data(data: tuple[DigitalTransition]):
        """Postprocess the pendulum raw data in order to calculate the period and
        the transit time for all the oscillations.

        Mind we skip the first edge if the polarity is wrong---i.e. the
        data acquisition is started with the flag covering the optical gate.
        """
        # Note when Edge is an IntEnum we can get rid of the value, here.
        if data[0].edge == Edge.RISING:
            logger.info("Wrong edge detected on the first transition, skipping it...")
            data = data[1:]
        return Pendulum._postprocess_data_smooth(data)

    def post_stop(self, run_control: RunControlBase) -> None:
        """Overloaded method.

        And here we are post-processing the raw data file to calculate the actual
        high-level quantities used in the analysis.
        """
        file_path = run_control.data_file_path()
        logger.info(f"Post-processing {file_path}...")
        with PacketFile(DigitalTransition).open(file_path) as input_file:
            data = input_file.read_all()
        oscillations = self.postprocess_data(data)
        file_path = Path(f"{run_control.output_file_path_base()}_data_proc.txt")
        logger.info(f"Writing output file {file_path}...")
        with open(file_path, "w", encoding=BALDAQUIN_ENCODING) as output_file:
            output_file.write(Oscillation.text_header(creator=self.NAME))
            for oscillation in oscillations:
                output_file.write(oscillation.to_text())
        logger.info("Done.")

    def process_packet(self, packet_data: bytes) -> AbstractPacket:
        """Overloaded method.
        """
        transition = DigitalTransition.unpack(packet_data)
        return transition


def main() -> None:
    """Main entry point.
    """
    bootstrap_window(AppMainWindow, PlasduinoRunControl(), Pendulum())


if __name__ == "__main__":
    main()



# Copyright (C) 2022--2024 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Silly application with Histogram display.
"""

import numpy as np
from aptapy.hist import Histogram1d

from baldaquin import silly
from baldaquin.__qt__ import QtWidgets
from baldaquin.gui import bootstrap_window
from baldaquin.pkt import AbstractPacket
from baldaquin.silly.common import (
    SillyConfiguration,
    SillyMainWindow,
    SillyPacket,
    SillyRunControl,
    SillyUserApplicationBase,
)


class MainWindow(SillyMainWindow):

    def __init__(self, parent: QtWidgets.QWidget = None) -> None:
        """Constructor.
        """
        super().__init__()
        self.hist_tab = self.add_plot_canvas_tab("PHA distribution")
        self.tab_widget.setCurrentWidget(self.hist_tab)

    def setup_user_application(self, user_application):
        """Overloaded method.
        """
        super().setup_user_application(user_application)
        self.hist_tab.register(user_application.pha_hist)


class SillyHist(SillyUserApplicationBase):

    """Simplest possible user application for testing purposes.
    """

    NAME = "Silly histogram display"
    CONFIGURATION_CLASS = SillyConfiguration
    CONFIGURATION_FILE_PATH = silly.SILLY_APP_CONFIG / "silly_hist.cfg"

    def __init__(self):
        """Overloaded constructor.
        """
        super().__init__()
        self.pha_hist = Histogram1d(np.linspace(800., 1200., 100), xlabel="PHA [ADC counts]")

    def process_packet(self, packet_data: bytes) -> AbstractPacket:
        """Dumb data processing routine---print out the actual event.
        """
        packet = SillyPacket.unpack(packet_data)
        self.pha_hist.fill(packet.pha)
        return packet


def main() -> None:
    """Main entry point.
    """
    bootstrap_window(MainWindow, SillyRunControl(), SillyHist())


if __name__ == "__main__":
    main()



# Copyright (C) 2022--2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Basic run control structure.
"""

import json
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

from . import __version__
from .__qt__ import QtCore
from .app import UserApplicationBase
from .config import UserApplicationConfiguration
from .env import BALDAQUIN_ENCODING, config_folder_path, data_folder_path
from .event import PacketStatistics
from .logging_ import logger, setup_logger, start_file_logging
from .timeline import Timeline, Timestamp


class FsmState(Enum):

    """Enum for the run control finite state machine possible states.
    """

    RESET = "Reset"
    STOPPED = "Stopped"
    RUNNING = "Running"
    PAUSED = "Paused"


class InvalidFsmTransitionError(RuntimeError):

    """RuntimeError subclass to signal an invalid FSM transition.
    """

    def __init__(self, src, dest):
        """Constructor.
        """
        super().__init__(f"Invalid FSM transition {src.name} -> {dest.name}.")


class FiniteStateMachineLogic:

    """Class encapsulating the basic logic of the run control finite-state machine.
    """

    def __init__(self) -> None:
        """Constructor.
        """
        self._state = FsmState.RESET

    def state(self) -> FsmState:
        """Return the state of the FSM.
        """
        return self._state

    def set_state(self, state: FsmState) -> None:
        """Set the state of the FSM.
        """
        self._state = state

    def is_reset(self) -> bool:
        """Return True if the run control is reset.
        """
        return self._state == FsmState.RESET

    def is_stopped(self) -> bool:
        """Return True if the run control is stopped.
        """
        return self._state == FsmState.STOPPED

    def is_running(self) -> bool:
        """Return True if the run control is running.
        """
        return self._state == FsmState.RUNNING

    def is_paused(self) -> bool:
        """Return True if the run control is paused.
        """
        return self._state == FsmState.PAUSED

    def setup(self) -> None:
        """Method called in the ``RESET`` -> ``STOPPED`` transition.
        """
        raise NotImplementedError

    def teardown(self) -> None:
        """Method called in the ``STOPPED`` -> ``RESET`` transition.
        """
        raise NotImplementedError

    def start_run(self) -> None:
        """Method called in the ``STOPPED`` -> ``RUNNING`` transition.
        """
        raise NotImplementedError

    def stop_run(self) -> None:
        """Method called in the ``RUNNING`` -> ``STOPPED`` transition.
        """
        raise NotImplementedError

    def pause(self) -> None:
        """Method called in the ``RUNNING`` -> ``PAUSED`` transition.
        """
        raise NotImplementedError

    def resume(self) -> None:
        """Method called in the ``PAUSED -> ``RUNNING`` transition.
        """
        raise NotImplementedError

    def stop(self) -> None:
        """Method called in the ``PAUSED`` -> ``STOPPED`` transition.
        """
        raise NotImplementedError

    def force_reset(self) -> None:
        """Go through all the necessary transitions to bring the FSM from the current
        state all the way to the ``RESET`` state.

        This is used, e.g., when the GUI is closed abruptly from the top-right cross
        and we still need to stop the data acquisition and put the hardware in a
        secure state before we actually shut everything down.
        """
        # If the run control is in either the RUNNING or PAUSED states, we first
        # need to transition to STOPPED.
        if self.is_running():
            self.stop_run()
        elif self.is_paused():
            self.stop()
        # At this point the status is STOPPED, and we just have to call teardown().
        if self.is_stopped():
            self.teardown()
        # And if the thing was in the RESET state to start with, then we're good.

    def set_reset(self) -> None:
        """Set the FSM in the ``RESET`` state.

        An :class:`InvalidFsmTransitionError <baldaquin.runctrl.InvalidFsmTransitionError>`
        exception is raised if the FSM is not in the ``STOPPED`` state.
        """
        target_state = FsmState.RESET
        if self.is_stopped():
            self.teardown()
        else:
            raise InvalidFsmTransitionError(self._state, target_state)
        self.set_state(target_state)

    def set_stopped(self) -> None:
        """Set the FSM in the ``STOPPED`` state.

        An :class:`InvalidFsmTransitionError <baldaquin.runctrl.InvalidFsmTransitionError>`
        exception is raised if the FSM is not in either the ``RESET``, ``RUNNING``
        or ``PAUSED`` state.
        """
        target_state = FsmState.STOPPED
        if self.is_reset():
            self.setup()
        elif self.is_running():
            self.stop_run()
        elif self.is_paused():
            self.stop()
        else:
            raise InvalidFsmTransitionError(self._state, target_state)
        self.set_state(target_state)

    def set_running(self) -> None:
        """Set the FSM in the ``RUNNING`` state.

        An :class:`InvalidFsmTransitionError <baldaquin.runctrl.InvalidFsmTransitionError>`
        exception is raised if the FSM is not in either the ``STOPPED`` or ``PAUSED`` state.
        """
        target_state = FsmState.RUNNING
        if self.is_stopped():
            self.start_run()
        elif self.is_paused():
            self.resume()
        else:
            raise InvalidFsmTransitionError(self._state, target_state)
        self.set_state(target_state)

    def set_paused(self) -> None:
        """Set the FSM in the ``PAUSED`` state.

        An :class:`InvalidFsmTransitionError <baldaquin.runctrl.InvalidFsmTransitionError>`
        exception is raised if the FSM is not in either the ``RUNNING`` state.
        """
        target_state = FsmState.PAUSED
        if self.is_running():
            self.pause()
        else:
            raise InvalidFsmTransitionError(self._state, target_state)
        self.set_state(target_state)


class FiniteStateMachineBase(QtCore.QObject, FiniteStateMachineLogic):

    """Definition of the finite-state machine (FSM) underlying the run control.

    This is inheriting from FiniteStateMachineLogic and overloading the set_state()
    hook, so that a state_changed signal is emitted whenever the state is changed.
    (Note that, in order to do this, we also have to overload the constructor in
    order for the underlying QObject structure to be properly initialized.)
    """

    # pylint: disable=c-extension-no-member, abstract-method
    state_changed = QtCore.Signal(FsmState)

    def __init__(self) -> None:
        """Overloaded constructor.
        """
        QtCore.QObject.__init__(self)
        FiniteStateMachineLogic.__init__(self)

    def set_state(self, state: FsmState) -> None:
        """Set the state of the FSM and emit a ``state_changed()`` signal with the
        proper state after the change.
        """
        self._state = state
        self.state_changed.emit(self._state)


class AppNotLoadedError(RuntimeError):

    """RuntimeError subclass to signal that the run control has no user application loaded.
    """

    def __init__(self):
        """Constructor.
        """
        super().__init__("User application not loaded.")


@dataclass
class RunReport:

    """Small container class describing a run report.
    """

    baldaquin_version: str
    test_stand_id: int
    run_id: int
    start_timestamp: Timestamp
    stop_timestamp: Timestamp
    project_name: str
    application_name: str
    statistics: PacketStatistics

    _VERSION = 1
    _VERSION_FIELD_NAME = "report_version"

    def to_dict(self):
        """Serialization.
        """
        _dict = {self._VERSION_FIELD_NAME: self._VERSION}
        _dict.update(self.__dict__)
        for key in ("start_timestamp", "stop_timestamp", "statistics"):
            _dict[key] = _dict[key].to_dict()
        return _dict

    @classmethod
    def from_dict(cls, **kwargs) -> "RunReport":
        """Deserialization.
        """
        _ = kwargs.pop(cls._VERSION_FIELD_NAME)
        for key in ("start_timestamp", "stop_timestamp"):
            kwargs.update({key: Timestamp.from_dict(**kwargs[key])})
        for key in ("statistics", ):
            kwargs.update({key: PacketStatistics.from_dict(**kwargs[key])})
        return cls(**kwargs)

    def dumps(self, indent: int = 4) -> str:
        """Return a text representation of the object in json format.
        """
        return json.dumps(self.to_dict(), indent=indent)

    def save(self, file_path: str) -> None:
        """Save the report to file.
        """
        logger.info(f"Writing run report to {file_path}...")
        with open(file_path, "w", encoding=BALDAQUIN_ENCODING) as output_file:
            output_file.write(self.dumps())

    @classmethod
    def load(cls, file_path):
        """Load the report from file.
        """
        logger.info(f"Loading run report from {file_path}...")
        with open(file_path, encoding=BALDAQUIN_ENCODING) as input_file:
            return cls.from_dict(**json.load(input_file))


class RunControlBase(FiniteStateMachineBase):

    """Run control class.

    Derived classes need to set the ``_PROJECT_NAME`` class member (this controls
    the placement of the output files) and, optionally ``_DEFAULT_REFRESH_INTERVAL``
    as well.

    Arguments
    ---------
    refresh_interval : int
        The timeout interval (in ms) for the underlying refresh QTimer object
        updating the information on the control GUI as the data taking proceeds.
    """

    # pylint: disable=c-extension-no-member, too-many-instance-attributes, too-many-public-methods
    _PROJECT_NAME = None
    _DEFAULT_REFRESH_INTERVAL = 750

    run_id_changed = QtCore.Signal(int)
    user_application_loaded = QtCore.Signal(UserApplicationBase)
    uptime_updated = QtCore.Signal(float)
    event_handler_stats_updated = QtCore.Signal(PacketStatistics, float)

    def __init__(self, refresh_interval: int = _DEFAULT_REFRESH_INTERVAL) -> None:
        """Constructor.
        """
        if self._PROJECT_NAME is None:
            msg = f"{self.__class__.__name__} needs to be subclassed, and _PROJECT_NAME set."
            raise RuntimeError(msg)
        super().__init__()
        self._test_stand_id = self._read_test_stand_id()
        self._run_id = self._read_run_id()
        self.timeline = Timeline()
        self.start_timestamp = None
        self.stop_timestamp = None
        self._user_application = None
        self._log_file_handler_id = None
        self._update_timer = QtCore.QTimer()
        self.set_refresh_interval(refresh_interval)
        self._update_timer.timeout.connect(self.update_stats)

    def test_stand_id(self) -> int:
        """Return the test-stand ID.
        """
        return self._test_stand_id

    def run_id(self) -> int:
        """Return the current run ID.
        """
        return self._run_id

    def set_refresh_interval(self, refresh_interval: int) -> None:
        """Set the timeout for the underlying refresh QTimer object.

        Arguments
        ---------
        refresh_interval : int
            The refresh interval in ms.
        """
        self._update_timer.setInterval(refresh_interval)

    def _config_file_path(self, file_name: str) -> Path:
        """Return the path to a generic configuration file.

        Arguments
        ---------
        file_name : str
            The file name.
        """
        return config_folder_path(self._PROJECT_NAME) / file_name

    def _test_stand_id_file_path(self) -> Path:
        """Return the path to the configuration file holding the test stand id.
        """
        return self._config_file_path("test_stand.cfg")

    def _run_id_file_path(self) -> Path:
        """Return the path to the configuration file holding the run id.
        """
        return self._config_file_path("run.cfg")

    def _file_name_base(self, label: str = None, extension: str = None) -> str:
        """Generic function implementing a file name factory, given the
        test stand and the run ID.

        Arguments
        ---------
        label : str
            A text label to attach to the file name.

        extension : str
            The file extension
        """
        file_name = f"{self._test_stand_id:04d}_{self._run_id:06d}"
        if label is not None:
            file_name = f"{file_name}_{label}"
        if extension is not None:
            file_name = f"{file_name}.{extension}"
        return file_name

    def data_folder_path(self) -> Path:
        """Return the path to the data folder for the current run.
        """
        return data_folder_path(self._PROJECT_NAME) / self._file_name_base()

    def output_file_path_base(self) -> Path:
        """Return the base pattern for all the output files.

        This is use to pass the message about where to write output files to
        user applications.
        """
        return self.data_folder_path() / self._file_name_base()

    def data_file_name(self) -> str:
        """Return the file name for the current data file.

        Note that RunControlBase subclasses can overload this if a different
        naming convention is desired.
        """
        return self._file_name_base("data", "dat")

    def data_file_path(self) -> Path:
        """Return the path to the current data file.
        """
        return self.data_folder_path() / self.data_file_name()

    def log_file_name(self):
        """Return the file name for the current log file.

        Note that RunControlBase subclasses can overload this if a different
        naming convention is desired.
        """
        return self._file_name_base("run", "log")

    def log_file_path(self) -> Path:
        """Return the path to the current log file.
        """
        return self.data_folder_path() / self.log_file_name()

    def config_file_name(self) -> str:
        """Return the file name for the current configuration.
        """
        return self._file_name_base("config", "json")

    def config_file_path(self) -> Path:
        """Return the path to the current configuration.
        """
        return self.data_folder_path() / self.config_file_name()

    def report_file_name(self) -> str:
        """Return the file name for the current run report.
        """
        return self._file_name_base("report", "json")

    def report_file_path(self) -> Path:
        """Return the path to the current run report.
        """
        return self.data_folder_path() / self.report_file_name()

    @staticmethod
    def _read_config_file(file_path: Path, default: int) -> int:
        """Read a single integer value from a given configuration file.

        If the file is not found, a new one is created, holding the default value,
        and the latter is returned.

        Arguments
        ---------
        file_path : Path
            The path to the configuration file.

        default : int
            The default value, to be used if the file is not found.
        """
        if not file_path.exists():
            logger.warning(f"Configuration file {file_path} not found, creating one...")
            RunControlBase._write_config_file(file_path, default)
            return default
        logger.info(f"Reading configuration file {file_path}...")
        value = int(file_path.read_text())
        logger.info(f"Done, {value} found.")
        return value

    @staticmethod
    def _write_config_file(file_path: Path, value: int) -> None:
        """Write a single integer value to a given configuration file.

        Arguments
        ---------
        file_path : Path
            The path to the configuration file.

        value : int
            The value to be written.
        """
        logger.info(f"Writing {value} to config file {file_path}...")
        file_path.write_text(f"{value}")

    def _read_test_stand_id(self, default: int = 101) -> int:
        """Read the test stand id from the proper configuration file.
        """
        return self._read_config_file(self._test_stand_id_file_path(), default)

    def _read_run_id(self) -> int:
        """Read the run ID from the proper configuration file.
        """
        return self._read_config_file(self._run_id_file_path(), 0)

    def _write_run_id(self) -> None:
        """Write the current run ID to the proper configuration file.
        """
        self._write_config_file(self._run_id_file_path(), self._run_id)

    def _increment_run_id(self) -> None:
        """Increment the run ID by one unit and update the corresponding
        configuration file.
        """
        self._run_id += 1
        self.run_id_changed.emit(self._run_id)
        self._write_run_id()

    def _create_data_folder(self) -> None:
        """Create the folder for the output data.
        """
        folder_path = self.data_folder_path()
        logger.info(f"Creating output data folder {folder_path}")
        Path.mkdir(folder_path)

    def elapsed_time(self) -> float:
        """Return the elapsed time.

        The precise semantics of this function is:

        * is the run control is either running or paused, return the elapsed time
          since the start of the run;
        * if both the start and the stop timestamps are not None, then return the
          total elapsed time in the last run;
        * if both of the above fail, then return None.
        """
        if self.is_running() or self.is_paused():
            return self.timeline.latch() - self.start_timestamp
        try:
            return self.stop_timestamp - self.start_timestamp
        except TypeError:
            return None

    def update_stats(self):
        """Signal the proper updates to the run statistics.
        """
        elapsed_time = self.elapsed_time()
        statistics = self._user_application.event_handler.statistics()
        try:
            event_rate = statistics.packets_processed / elapsed_time
        except TypeError:
            event_rate = 0.
        self.uptime_updated.emit(elapsed_time)
        self.event_handler_stats_updated.emit(statistics, event_rate)

    def write_run_report(self) -> None:
        """Write an end-of-run report in the output folder.
        """
        report = RunReport(__version__, self._test_stand_id, self._run_id, self.start_timestamp,
                           self.stop_timestamp, self._PROJECT_NAME,
                           self._user_application.__class__.__name__,
                           self._user_application.event_handler.statistics())
        report.save(self.report_file_path())

    def load_user_application(self, user_application: UserApplicationBase) -> None:
        """Set the user application to be run.
        """
        logger.info("Loading user application...")
        if not self.is_reset():
            raise RuntimeError(f"Cannot load a user application in the {self.state().name} state")
        if not isinstance(user_application, UserApplicationBase):
            raise RuntimeError(f"Invalid user application of type {type(user_application)}")
        self._user_application = user_application
        # Mind we want to set the state to STOPPED before we emit the user_application_loaded()
        # signal, in order to avoid triggering invalid transtions downstream.
        self.set_stopped()
        self.user_application_loaded.emit(user_application)

    def _check_user_application(self) -> None:
        """Make sure we have a valid use application loaded, and raise an
        AppNotLoadedError if that is not the case.
        """
        if self._user_application is None:
            raise AppNotLoadedError

    def configure_user_application(self, configuration: UserApplicationConfiguration) -> None:
        """Apply a given configuration to the current user application.
        """
        self._check_user_application()
        logger.info("Configuring user application...")
        self._user_application.apply_configuration(configuration)

    def setup(self) -> None:
        """Overloaded method.
        """
        self._check_user_application()
        self._user_application.setup()

    def teardown(self) -> None:
        """Overloaded method.
        """
        self._check_user_application()
        self._user_application.teardown()

    def start_run(self) -> None:
        """Overloaded method.
        """
        # Note that configure_user_application() has been called before this.
        # We should think carefully about whether we're doing the right thing...
        self._check_user_application()
        self._increment_run_id()
        self._create_data_folder()
        # Apply all the necessary configurations. Note this needs to happen after
        # the run ID has been incremented, so that the log file points to the
        # right place.
        configuration = self._user_application.configuration
        logger.info(f"Applying configuration...\n{configuration}")
        logger.info("Configuring logging...")
        section = configuration.logging_section()
        setup_logger(section.value("terminal_level"))
        self._log_file_handler_id = start_file_logging(self.log_file_path(),
                                                       section.value("file_level"))
        logger.info("Configuring packet buffering...")
        section = configuration.buffering_section()
        flush_size = section.value("flush_size")
        flush_timeout = section.value("flush_timeout")
        self._user_application.event_handler.configure_buffer(flush_size, flush_timeout)
        configuration.save(self.config_file_path())
        # Configuration applied and written, we might move on.
        self.start_timestamp = self.timeline.latch()
        self.stop_timestamp = None
        logger.info(f"Run Control started on {self.start_timestamp}")
        self._user_application.event_handler.set_primary_sink(self.data_file_path())
        self._user_application.pre_start(self)
        self._user_application.start_run()
        self._update_timer.start()
        self.update_stats()

    def stop_run(self) -> None:
        """Overloaded method.
        """
        self._check_user_application()
        self._update_timer.stop()
        self._user_application.stop_run()
        self._user_application.post_stop(self)
        self.stop_timestamp = self.timeline.latch()
        logger.info(f"Run Control stopped on {self.stop_timestamp}")
        logger.info(f"Total elapsed time: {self.elapsed_time():6f} s.")
        if self._log_file_handler_id is not None:
            logger.remove(self._log_file_handler_id)
        self._log_file_handler_id = None
        self.write_run_report()
        self.update_stats()

    def pause(self) -> None:
        """Overloaded method.
        """
        self._check_user_application()
        self._user_application.pause()

    def resume(self) -> None:
        """Overloaded method.
        """
        self._check_user_application()
        self._user_application.resume()

    def stop(self) -> None:
        """Overloaded method.
        """
        self.stop_run()



# Copyright (C) 2024--25 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Plasduino pendulum viewer application.
"""

from pathlib import Path

from baldaquin import plasduino
from baldaquin.__qt__ import QtWidgets
from baldaquin.buf import WriteMode
from baldaquin.gui import MainWindow, SimpleControlBar, bootstrap_window
from baldaquin.pkt import AbstractPacket, packetclass
from baldaquin.plasduino import PLASDUINO_APP_CONFIG
from baldaquin.plasduino.common import (
    PlasduinoAnalogConfiguration,
    PlasduinoAnalogEventHandler,
    PlasduinoAnalogUserApplicationBase,
    PlasduinoRunControl,
)
from baldaquin.plasduino.protocol import AnalogReadout
from baldaquin.plasduino.shields import Lab1
from baldaquin.runctrl import RunControlBase


class AppMainWindow(MainWindow):

    """Application graphical user interface.
    """

    _PROJECT_NAME = plasduino.PROJECT_NAME
    _CONTROL_BAR_CLASS = SimpleControlBar

    def __init__(self, parent: QtWidgets.QWidget = None) -> None:
        """Constructor.
        """
        super().__init__()
        self.strip_chart_tab = self.add_plot_canvas_tab("Strip charts", update_interval=100)

    def setup_user_application(self, user_application):
        """Overloaded method.
        """
        super().setup_user_application(user_application)
        self.strip_chart_tab.register(*user_application.strip_chart_dict.values())


@packetclass
class PositionReadout(AnalogReadout):

    """Specialized class inheriting from ``AnalogReadout`` describing a position
    readout---this is just changing the label for the text otuput.
    """

    OUTPUT_HEADERS = ("Pin number", "Time [s]", "Position [a. u.]")


class PendulumView(PlasduinoAnalogUserApplicationBase):

    """Simplest possible user application for testing purposes.
    """

    NAME = "Pendulum View"
    CONFIGURATION_CLASS = PlasduinoAnalogConfiguration
    CONFIGURATION_FILE_PATH = PLASDUINO_APP_CONFIG / "plasduino_pendulumview.cfg"
    EVENT_HANDLER_CLASS = PlasduinoAnalogEventHandler
    _PINS = Lab1.PENDVIEW_PINS
    _SAMPLING_INTERVAL = 50

    def __init__(self) -> None:
        """Overloaded Constructor.
        """
        super().__init__()
        self.strip_chart_dict = self.create_strip_charts(self._PINS, ylabel="Position [ADC counts]")

    def configure(self) -> None:
        """Overloaded method.
        """
        max_length = self.configuration.application_section().value("strip_chart_max_length")
        for chart in self.strip_chart_dict.values():
            chart.set_max_length(max_length)

    def pre_start(self, run_control: RunControlBase) -> None:
        """Overloaded method.
        """
        file_path = Path(f"{run_control.output_file_path_base()}_data.txt")
        self.event_handler.add_custom_sink(file_path, WriteMode.TEXT, PositionReadout.to_text,
                                           PositionReadout.text_header(creator=self.NAME))

    def process_packet(self, packet_data: bytes) -> AbstractPacket:
        """Overloaded method.
        """
        readout = PositionReadout.unpack(packet_data)
        self.strip_chart_dict[readout.pin_number].put(readout.seconds, readout.adc_value)
        return readout


def main() -> None:
    """Main entry point.
    """
    bootstrap_window(AppMainWindow, PlasduinoRunControl(), PendulumView())


if __name__ == "__main__":
    main()



# Copyright (C) 2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for the plasduino project.
"""

import importlib
import sys

import numpy as np
import pytest

from baldaquin.env import BALDAQUIN_SOURCE
from baldaquin.pkt import PacketFile
from baldaquin.plasduino.protocol import AnalogReadout, DigitalTransition
from baldaquin.plt_ import plt, setup_gca

PENDULUM_RUN = "396"


def test_protocol():
    """Test the protocol.
    """
    _ = AnalogReadout(0xa2, 1, 1000, 255)
    _ = DigitalTransition(0xa1, 1, 1000000)


@pytest.mark.skip
def test_pendulum_process(test_data_path):
    """Test the pendulum post-processing code.

    This was done to debug issue https://github.com/lucabaldini/baldaquin/issues/50
    """
    # pylint: disable=protected-access
    sys.path.append(str(BALDAQUIN_SOURCE / "plasduino" / "apps"))
    sys.dont_write_bytecode = True
    pendulum = importlib.import_module("plasduino_pendulum")
    sys.dont_write_bytecode = False
    file_path = test_data_path(f"0101_000{PENDULUM_RUN}", f"0101_000{PENDULUM_RUN}_data.dat")
    with PacketFile(DigitalTransition).open(file_path) as input_file:
        data = input_file.read_all()

    # Post-process with the simple method.
    oscillations = pendulum.Pendulum._postprocess_data_simple(data)
    simple_time = np.array([oscillation.average_time for oscillation in oscillations])
    simple_period = np.array([oscillation.period for oscillation in oscillations])
    simple_transit_time = np.array([oscillation.transit_time for oscillation in oscillations])

    # Post-process with the smoothed method.
    oscillations = pendulum.Pendulum._postprocess_data_smooth(data)
    smooth_time = np.array([oscillation.average_time for oscillation in oscillations])
    smooth_period = np.array([oscillation.period for oscillation in oscillations])
    smooth_transit_time = np.array([oscillation.transit_time for oscillation in oscillations])

    plt.figure("Simple processing: period")
    plt.plot(simple_time, simple_period, label="Simple")
    plt.plot(smooth_time, smooth_period, label="Smooth")
    setup_gca(xlabel="Time [s]", ylabel="Period [s]", grids=True, legend=True)
    plt.figure("Simple processing: transit time")
    plt.plot(simple_time, simple_transit_time, label="Simple")
    plt.plot(smooth_time, smooth_transit_time, label="Smooth")
    setup_gca(xlabel="Time [s]", ylabel="Period [s]", grids=True, legend=True)


def transit_velocity(transit_time: np.array, pendulum_length: float, gate_distance: float,
                     flag_width: float) -> np.array:
    """Calculate the average transit velocity based on the transit time.

    Arguments
    ---------
    transit_time : array_like
        The flag transit time.

    pendulum_length : float
        The pendulum length.

    gate_distance : float
        The distance between the suspension point an the optical gate.

    flag_width : float
        The width of the measuring flag.
    """
    return flag_width / transit_time * pendulum_length / gate_distance


def period_model(theta, T0):
    """Fitting model for the period as a function of the amplitude.
    """
    # pylint: disable=invalid-name
    return T0 * (1. + 1. / 16. * theta**2 + 11. / 3072. * theta**4. + 173. / 737280. * theta**6.)


def test_pendulum_sequence(test_data_path):
    """Draw the signal sequence.
    """
    file_path = test_data_path(f"0101_000{PENDULUM_RUN}", f"0101_000{PENDULUM_RUN}_data.dat")
    with PacketFile(DigitalTransition).open(file_path) as input_file:
        data = input_file.read_all()
    sequence = data[-11:-1]
    t0 = sequence[0].microseconds
    x = []
    y = []
    for transition in sequence:
        t = (transition.microseconds - t0) / 1000.
        x += [t, t]
        if transition.edge == 1:
            y += [1, 0]
        else:
            y += [0, 1]
    plt.figure("Initial sequence")
    plt.plot(x, y)
    setup_gca(ymax=1.1, xlabel="Time [ms]", ylabel="Status (high = occulted)")


def test_pendulum_plot(test_data_path):
    """Test a data file taken with the pendulum.
    """
    # pylint: disable=invalid-name
    g = 9.81
    mass = 0.330
    pendulum_length = 1.110
    gate_distance = 1.151
    flag_width = 0.0194
    T0 = 2. * np.pi * np.sqrt(pendulum_length / g)
    # optical_gate_width = 0.001

    file_path = test_data_path(f"0101_000{PENDULUM_RUN}", f"0101_000{PENDULUM_RUN}_data_proc.txt")
    time_, period, transit_time = np.loadtxt(file_path, delimiter=",", unpack=True)
    velocity = transit_velocity(transit_time, pendulum_length, gate_distance, flag_width)

    # Correct the period for the width of the optical gate!
    # period -= optical_gate_width / velocity

    amplitude = np.arccos(1. - velocity**2. / 2. / g / pendulum_length)
    energy = 0.5 * mass * velocity**2.
    energy_loss = np.diff(energy) / (0.5 * (energy[:-1] + energy[1:]))

    plt.figure("Period")
    plt.plot(time_, period, "o")
    setup_gca(xlabel="Time [s]", ylabel="Period [s]", grids=True)

    # plt.figure("Transit time")
    # plt.plot(time_, transit_time, "o")
    # setup_gca(xlabel="Time [s]", ylabel="Transit time [s]", grids=True)

    plt.figure("Amplitude")
    plt.plot(amplitude, period, "o")
    plt.plot(amplitude, period_model(amplitude, T0))
    setup_gca(xlabel="Amplitude [rad]", ylabel="Period [s]", grids=True)

    plt.figure("Amplitude residuals")
    plt.plot(amplitude, period - period_model(amplitude, T0), "o")
    setup_gca(xlabel="Amplitude [rad]", ylabel="Period residuals [s]", grids=True)

    plt.figure("Energy")
    plt.plot(time_, energy, "o")
    setup_gca(xlabel="Period [s]", ylabel="Energy [J]", grids=True)

    plt.figure("Energy loss")
    plt.plot(time_[1:], energy_loss * 100., "o")
    setup_gca(xlabel="Period [s]", ylabel="Fractional energy loss [%]", grids=True)



# Copyright (C) 2022--2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Custom widgets for the user interface.
"""

import sys
from enum import Enum
from pathlib import Path
from typing import Any

# This needs to stay *after* the from baldaquin.__qt__ import, in order for the
# matplotlib monkeypatching to happen in time.
# pylint: disable=no-name-in-module, wrong-import-order, ungrouped-imports, import-error
from matplotlib.backends.backend_qtagg import FigureCanvas
from matplotlib.figure import Figure

from .__qt__ import QtCore, QtGui, QtWidgets, exec_qapp
from .app import UserApplicationBase
from .config import ConfigurationParameter, UserApplicationConfiguration
from .env import BALDAQUIN_ICONS, BALDAQUIN_SKINS
from .pkt import PacketStatistics
from .runctrl import FiniteStateMachineLogic, FsmState, RunControlBase

# pylint: disable=too-many-lines


_DEFAULT_ICON_CATALOG = "material3"
_DEFAULT_ICON_STYLE = "FILL0_wght400_GRAD0_opsz48"


def _icon_file_path(name: str, catalog: str = _DEFAULT_ICON_CATALOG, fmt: str = "svg",
                    style: str = _DEFAULT_ICON_STYLE) -> str:
    """Return the path to a given icon file.

    Note this is returning a string, rather than a Path object, as the function
    is mainly used in Qt calls, where Path objects are not readily understood.
    """
    file_name = f"{name}_{style}.{fmt}"
    return f"{BALDAQUIN_ICONS / catalog / file_name}"


def load_icon(name: str, catalog: str = _DEFAULT_ICON_CATALOG, fmt: str = "svg",
              style: str = _DEFAULT_ICON_STYLE) -> QtGui.QIcon:
    """Load an icon from the given catalog.
    """
    return QtGui.QIcon(_icon_file_path(name, catalog, fmt, style))


def stylesheet_file_path(name: str = "default") -> Path:
    """Return the path to a given stylesheet file.
    """
    file_name = f"{name}.qss"
    return BALDAQUIN_SKINS / file_name


class Button(QtWidgets.QPushButton):

    """Small wrapper around the QtWidgets.QPushButton class.

    Arguments
    ---------
    icon_name : str
        The name of the icon to be associated to the button.

    size : int
        The button size.

    icon_size : int
        The icon  size.

    tooltip : str, optional
        An optional tooltip to be associated to the button.
    """

    # pylint: disable=too-few-public-methods

    DEFAULT_SIZE: int = 40
    DEFAULT_ICON_SIZE: int = 25

    def __init__(self, icon_name: str, tooltip: str = None, size: int = DEFAULT_SIZE,
                 icon_size: int = DEFAULT_ICON_SIZE) -> None:
        """Constructor.
        """
        super().__init__()
        self.setFixedSize(size, size)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        self.set_icon(icon_name, icon_size)
        if tooltip:
            self.setToolTip(tooltip)

    def set_icon(self, icon_name: str, icon_size: int = DEFAULT_ICON_SIZE) -> None:
        """Set the button icon.

        Note that when icon_name is an Enum instance we automatically get the
        Enum value.
        """
        if isinstance(icon_name, Enum):
            icon_name = icon_name.value
        self.setIcon(load_icon(icon_name))
        self.setIconSize(QtCore.QSize(icon_size, icon_size))


class DataWidgetBase(QtWidgets.QWidget):

    """Base class for a widget holding a single piece of information.

    This is nothing more than a pair of widget---a QLabel holding a title and
    a generic widget holding some data---arranged in a vertical layout.
    This is an abstract class, and specialized subclasses are provided below
    to show or edit data in different fashions, representing the basic building
    block to construct complex user interfaces.

    Derived classes must override, at the very minimum, the ``VALUE_WIDGET_CLASS``
    class member and the ``current_value()`` and ``set_value()`` class methods;
    overloading ``setup()`` is optional.

    Note that the constructor of this base class calls the ``setObjectName()``
    method for both widgets (with different arguments) to make it easier to
    stylesheet the application downstream.
    """

    VALUE_WIDGET_CLASS = None
    TITLE_WIDGET_NAME = "data_widget_title"
    VALUE_WIDGET_NAME = "data_widget_value"
    MISSING_VALUE_LABEL = "-"
    VALUE_WIDGET_HEIGHT = 28

    def __init__(self, name: str, title: str = None, value=None, units: str = None,
                 fmt: str = None, **kwargs) -> None:
        """Constructor
        """
        # pylint: disable=not-callable, too-many-arguments
        if title is None:
            title = name
        self.name = name
        self._units = units
        self._fmt = fmt
        super().__init__()
        self.setLayout(QtWidgets.QVBoxLayout())
        self.title_widget = QtWidgets.QLabel()
        self.title_widget.setObjectName(self.TITLE_WIDGET_NAME)
        self.value_widget = self.VALUE_WIDGET_CLASS()
        self.value_widget.setObjectName(self.VALUE_WIDGET_NAME)
        self.value_widget.setFixedHeight(self.VALUE_WIDGET_HEIGHT)
        self.layout().addWidget(self.title_widget)
        self.layout().addWidget(self.value_widget)
        self.set_title(title)
        self.setup(**kwargs)
        if value is not None:
            self.set_value(value)
        else:
            self.set_value(self.MISSING_VALUE_LABEL)

    def setup(self, **kwargs):
        """Do nothing post-construction hook that can be overloaded in derived classes.
        """

    def set_title(self, title: str) -> None:
        """Set the widget title.
        """
        self.title_widget.setText(title)

    def current_value(self) -> Any:
        """Return the current value hold by the configuration widget.
        """
        raise NotImplementedError

    def set_value(self, value: Any) -> None:
        """Set hook to be reimplemented by derived classes.
        """
        raise NotImplementedError


class DisplayWidget(DataWidgetBase):

    """Simple widget to display a single piece of information in a read-only fashion.
    """

    VALUE_WIDGET_CLASS = QtWidgets.QLabel

    def set_value(self, value: Any) -> None:
        """Overloaded method.
        """
        text = f'{value:{self._fmt if self._fmt else ""}} {self._units if self._units else ""}'
        self.value_widget.setText(text)

    def current_value(self) -> Any:
        """Return the value of the widget.
        """
        return self.text()


class CardWidget(QtWidgets.QFrame):

    """Small class implementing a (read-only) card.
    """

    _FIELD_ENUM: Enum = ()
    _VALUE_DICT = {}
    _KWARGS_DICT = {}

    def __init__(self, add_bottom_stretch: bool = True) -> None:
        """Constructor.
        """
        super().__init__()
        self.setLayout(QtWidgets.QVBoxLayout())
        self._widget_dict = {}
        for field in self._FIELD_ENUM:
            value = self._VALUE_DICT.get(field)
            kwargs = self._KWARGS_DICT.get(field, {})
            self.add(field.value, None, value, **kwargs)
        if add_bottom_stretch:
            self.add_bottom_stretch()

    def add_bottom_stretch(self) -> None:
        """Add a vertical stretch at the bottom of the layout.

        This is to ensure that the card item are "floating" in the card container
        without being expanded, and the net effect is that the items appear
        aligned to the top in the card. There might be a clever way to achieve
        this automatically in the loop filling the card, but I could not make it
        work, of not manually.
        """
        self.layout().addStretch()

    def add(self, *args, **kwargs) -> None:
        """Add an item to the card.

        Note the signature, here, is designed to match exactly that of the
        :class:`DisplayWidget <baldaquin.widget.DisplayWidget>` constructor.
        """
        widget = DisplayWidget(*args, **kwargs)
        if widget.name in self._widget_dict:
            raise RuntimeError(f'Duplicated card item name "{widget.name}"')
        self._widget_dict[widget.name] = widget
        self.layout().addWidget(widget)

    def set(self, name: str, value: Any):
        """Set the value for a specific card item (addressed by name).

        This will raise a KeyError if the card item does not exist.
        """
        if isinstance(name, Enum):
            name = name.value
        if value is None:
            value = DataWidgetBase.MISSING_VALUE_LABEL
        self._widget_dict[name].set_value(value)

    @staticmethod
    def sizeHint() -> QtCore.QSize:
        """Overloaded method defining the default size.
        """
        # pylint: disable=invalid-name
        return QtCore.QSize(200, 400)


class RunControlCardField(Enum):

    """Definition of the relevant fields for the run control card.

    This is done with the intent of making the process of updating the values of
    the card less error-prone.
    """

    PROJECT_NAME = "Project"
    USER_APPLICATION = "User application"
    TEST_STAND_ID = "Test stand"
    RUN_ID = "Run ID"
    STATE = "State"
    UPTIME = "Uptime"


class RunControlCard(CardWidget):

    """Specialized card to display the run control information.
    """
    _FIELD_ENUM = RunControlCardField
    _VALUE_DICT = {
        RunControlCardField.UPTIME: 0.
        }
    _KWARGS_DICT = {
        RunControlCardField.UPTIME: dict(units="s", fmt=".3f")
        }


class EventHandlerCardField(Enum):

    """Specialized card for the event handler.
    """

    FILE_PATH = "Path to the output file"
    BUFFER_CLASS = "Data buffer type"
    NUM_PACKETS_PROCESSED = "Number of packets processed"
    AVERAGE_EVENT_RATE = "Average event rate"
    NUM_PACKETS_WRITTEN = "Number of packets written to disk"
    NUM_BYTES_WRITTEN = "Number of bytes written to disk"


class EventHandlerCard(CardWidget):

    """Specialized card widget for the event handler.
    """

    _FIELD_ENUM = EventHandlerCardField
    _VALUE_DICT = {
        EventHandlerCardField.NUM_PACKETS_PROCESSED: 0,
        EventHandlerCardField.AVERAGE_EVENT_RATE: 0.,
        EventHandlerCardField.NUM_PACKETS_WRITTEN: 0,
        EventHandlerCardField.NUM_BYTES_WRITTEN: 0
        }
    _KWARGS_DICT = {
        EventHandlerCardField.AVERAGE_EVENT_RATE: dict(units="Hz", fmt=".3f")
        }

    def __init__(self, add_bottom_stretch: bool = True) -> None:
        """Overloaded constructor.
        """
        super().__init__(add_bottom_stretch)
        key = EventHandlerCardField.FILE_PATH.value
        widget = self._widget_dict[key].title_widget
        widget.linkActivated.connect(MainWindow.open_url)

    def set_file_path(self, file_path: Path) -> None:
        """Specialize method to set the path to the output file.

        This gets its own method as we are also adding a hyperlink to the output
        folder.
        """
        self.set(EventHandlerCardField.FILE_PATH, file_path)
        key = EventHandlerCardField.FILE_PATH.value
        widget = self._widget_dict[key].title_widget
        text = f'{key} (<a href="{file_path.parent}">link to folder</a>)'
        widget.setText(text)


class ParameterCheckBox(DataWidgetBase):

    """Check box data widget, mapping ``bool`` input parameters.
    """

    VALUE_WIDGET_CLASS = QtWidgets.QCheckBox

    def current_value(self) -> bool:
        """Overloaded method.
        """
        return self.value_widget.isChecked()

    def set_value(self, value) -> None:
        """Overloaded method.
        """
        self.value_widget.setChecked(value)


class ParameterSpinBox(DataWidgetBase):

    """Spin box data widget, mapping ``int`` input parameters.

    There's a few interesting tweaks, here, that we have to take into account.
    The minimum and maximum values that the widget can hold are defined by the
    signature of the underlying C++ class, which is limited to a 32-bit unsigned
    integers, that is [-2,147,483,648, 2,147,483,647], inclusive. Trying to set
    a minimum/maximum value outside of this range will result in an OverflowError.
    """

    VALUE_WIDGET_CLASS = QtWidgets.QSpinBox
    _RANGE_MINIMUM = -2147483648
    _RANGE_MAXIMUM = 2147483647

    def setup(self, **kwargs) -> None:
        """Overloaded method.
        """
        self.value_widget.setMinimum(kwargs.get("min", self._RANGE_MINIMUM))
        self.value_widget.setMaximum(kwargs.get("max", self._RANGE_MAXIMUM))
        self.value_widget.setSingleStep(kwargs.get("step", 1))
        if self._units is not None:
            self.value_widget.setSuffix(f" {self._units}")

    def current_value(self) -> int:
        """Overloaded method.
        """
        return self.value_widget.value()

    def set_value(self, value) -> None:
        """Overloaded method.
        """
        if value < self._RANGE_MINIMUM or value > self._RANGE_MAXIMUM:
            raise OverflowError(f"Value {value} is outside of allowed range for "
                                f"{self.__class__.__name__} ")
        self.value_widget.setValue(value)


class ParameterDoubleSpinBox(ParameterSpinBox):

    """Double spin box data widget, mapping ``float`` input parameters.
    """

    VALUE_WIDGET_CLASS = QtWidgets.QDoubleSpinBox

    def current_value(self) -> float:
        """Overloaded method.
        """
        return self.value_widget.value()


class ParameterLineEdit(DataWidgetBase):

    """Line edit data widget, mapping ``str`` input parameters when there is
    no ``choice`` constraint applied.
    """

    VALUE_WIDGET_CLASS = QtWidgets.QLineEdit

    def current_value(self) -> str:
        """Overloaded method.
        """
        return self.value_widget.text()

    def set_value(self, value) -> None:
        """Overloaded method.
        """
        self.value_widget.setText(value)


class ParameterComboBox(DataWidgetBase):

    """Combo box data widget, mapping ``str`` input parameters when there is
    a ``choice`` constraint applied.
    """

    VALUE_WIDGET_CLASS = QtWidgets.QComboBox

    def setup(self, **kwargs) -> None:
        """Overloaded method.
        """
        self.value_widget.addItems(kwargs.get("choices", ()))

    def current_value(self) -> str:
        """Overloaded method.
        """
        return self.value_widget.currentText()

    def set_value(self, value) -> None:
        """Overloaded method.
        """
        self.value_widget.setCurrentIndex(self.value_widget.findText(value))


class ConfigurationSectionWidget(QtWidgets.QGroupBox):

    """Basic widget to display and edit an instance of a
    :class:`baldaquin.config.UserApplicationConfiguration` subclass.
    """

    _WIDGET_NAME = "configuration_section_widget"

    def __init__(self, configuration_section=None) -> None:
        """Constructor.
        """
        super().__init__()
        self.setLayout(QtWidgets.QVBoxLayout())
        self.setObjectName(self._WIDGET_NAME)
        self._section_class = None
        self._widget_dict = {}
        if configuration_section:
            self.display(configuration_section)

    def display(self, configuration_section) -> None:
        """Display a given configuration.
        """
        self._section_class = configuration_section.__class__
        self.setTitle(configuration_section.TITLE)
        self._widget_dict = {}
        for param in configuration_section.values():
            widget = self.__param_widget(param)
            self._widget_dict[widget.name] = widget
            self.layout().addWidget(widget)
        self.layout().addStretch()

    @staticmethod
    def __param_widget(param: ConfigurationParameter) -> DataWidgetBase:
        """Turn a configuration parameter into the corresponding widget.

        The fact that we're using two different widgets for ``str`` input,
        depending on whether ``choices`` is set or not, makes this slightly more
        lengthy to code---we could have probably just used a plain dictionary,
        if that was not the case.
        """
        args = param.name, param.intent, param.value, param.units, param.fmt
        kwargs = param.constraints
        type_ = param.type
        if type_ is bool:
            return ParameterCheckBox(*args, **kwargs)
        if type_ is int:
            return ParameterSpinBox(*args, **kwargs)
        if type_ is float:
            return ParameterDoubleSpinBox(*args, **kwargs)
        if type_ is str:
            if "choices" in kwargs:
                return ParameterComboBox(*args, **kwargs)
            return ParameterLineEdit(*args, **kwargs)
        raise RuntimeError(f"Unknown parameter type {type_}")

    def current_value(self, name: str) -> Any:
        """Retrieve the current value for a given key.
        """
        return self._widget_dict[name].current_value()

    def set_value(self, name: str, value: Any) -> None:
        """Set the value for a specific parameter (addressed by name).

        This will raise a KeyError if the widget does not exist.
        """
        self._widget_dict[name].set_value(value)

    @staticmethod
    def sizeHint() -> QtCore.QSize:
        """Overloaded method defining the default size.
        """
        # pylint: disable=invalid-name
        return QtCore.QSize(400, 400)

    def current_configuration_section(self):
        """Return the current configuration displayed in the widget.
        """
        section = self._section_class()
        for key, widget in self._widget_dict.items():
            section.set_value(key, widget.current_value())
        return section


class SettingsConfigurationWidget(QtWidgets.QWidget):

    """Composite widget to display the settings configuration.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self) -> None:
        """Constructor.
        """
        super().__init__()
        self.setLayout(QtWidgets.QGridLayout())
        self.layout().setColumnStretch(0, 1)
        self.layout().setColumnStretch(1, 1)
        self.logging_widget = None
        self.buffering_widget = None
        self.multicast_widget = None

    def display(self, configuration) -> None:
        """Display a given configuration.
        """
        self.logging_widget = ConfigurationSectionWidget(configuration.logging_section())
        self.layout().addWidget(self.logging_widget, 0, 0)
        self.buffering_widget = ConfigurationSectionWidget(configuration.buffering_section())
        self.layout().addWidget(self.buffering_widget, 0, 1)
        self.multicast_widget = ConfigurationSectionWidget(configuration.multicast_section())
        self.layout().addWidget(self.multicast_widget, 1, 0)
        # This is disabled until we have a proper multicast functionality in place.
        self.multicast_widget.setEnabled(False)


class PlotCanvasWidget(FigureCanvas):

    """Custom widget to display a matplotlib canvas.

    Arguments
    ---------
    update_interval : int
        The target interval for the update timer in ms.

    kwargs : dict
        The keyword arguments to be passed to the subplot() call.
    """

    def __init__(self, update_interval: int, **kwargs) -> None:
        """Constructor.
        """
        super().__init__(Figure())
        self.axes = self.figure.subplots(**kwargs)
        self._update_timer = QtCore.QTimer()
        self._update_timer.setInterval(update_interval)
        self._update_timer.timeout.connect(self._update)
        self._plot_list = []

    def register(self, *plots) -> None:
        """Register the underlying plots.

        All the plots passed as an argument here get passed to the ``_plot_list``
        class member, and get to be updated in the proper slot.
        """
        self._plot_list += plots
        self._update()

    def _update(self):
        """Update all the registered plots.
        """
        self.axes.clear()
        for plot in self._plot_list:
            plot.plot(self.axes)
        self.axes.set_autoscale_on(True)
        self.axes.legend(loc="upper left")
        self.axes.figure.canvas.draw()

    def start_updating(self) -> None:
        """Start the update timer.
        """
        self._update_timer.start()

    def stop_updating(self) -> None:
        """Stop the update timer and trigger one last, single-shot timeout()
        to capture the events collected after the stop.
        """
        self._update_timer.stop()
        self._update()


class ControlBarIcon(Enum):

    """Small enum for the icon of the control bar buttons.
    """

    TEARDOWN = "file_download"
    SETUP = "file_upload"
    START = "play_arrow"
    PAUSE = "pause"
    STOP = "stop"


class ControlBar(FiniteStateMachineLogic, QtWidgets.QFrame):

    """Control bar managing the run control.
    """

    # pylint: disable=c-extension-no-member
    set_reset_triggered = QtCore.Signal()
    set_stopped_triggered = QtCore.Signal()
    set_running_triggered = QtCore.Signal()
    set_paused_triggered = QtCore.Signal()

    def __init__(self, parent: QtWidgets.QWidget = None) -> None:
        """Constructor.
        """
        FiniteStateMachineLogic.__init__(self)
        QtWidgets.QFrame.__init__(self, parent)
        self.setLayout(QtWidgets.QHBoxLayout())
        # Create the necessary buttons.
        self.reset_button = self._add_button(ControlBarIcon.SETUP, "Setup/teardown")
        self.start_button = self._add_button(ControlBarIcon.START, "Start/pause")
        self.stop_button = self._add_button(ControlBarIcon.STOP, "Start/pause")
        # We start in the RESET state, where the start and stop buttons are disabled.
        self.start_button.setEnabled(False)
        self.stop_button.setEnabled(False)
        # Now: setup the connections.
        self.reset_button.clicked.connect(self.toggle_reset_button)
        self.start_button.clicked.connect(self.toggle_start_button)
        self.stop_button.clicked.connect(self.set_stopped)

    def _add_button(self, icon_name: ControlBarIcon, tooltip: str = None) -> Button:
        """Add a button to the control bar and return the Button object.
        """
        button = Button(icon_name, tooltip)
        self.layout().addWidget(button)
        return button

    def toggle_reset_button(self):
        """Toggle the reset button.
        """
        if self.is_reset():
            self.set_stopped()
        elif self.is_stopped():
            self.set_reset()

    def toggle_start_button(self):
        """Toggle the start button.
        """
        if self.is_running():
            self.set_paused()
        elif self.is_paused() or self.is_stopped():
            self.set_running()

    def setup(self) -> None:
        """Method called in the ``RESET`` -> ``STOPPED`` transition.
        """
        self.reset_button.set_icon(ControlBarIcon.TEARDOWN)
        self.start_button.setEnabled(True)

    def teardown(self) -> None:
        """Method called in the ``STOPPED`` -> ``RESET`` transition.
        """
        self.reset_button.set_icon(ControlBarIcon.SETUP)
        self.start_button.setEnabled(False)

    def start_run(self) -> None:
        """Method called in the ``STOPPED`` -> ``RUNNING`` transition.
        """
        self.reset_button.setEnabled(False)
        self.start_button.set_icon(ControlBarIcon.PAUSE)
        self.stop_button.setEnabled(True)

    def stop_run(self) -> None:
        """Method called in the ``RUNNING`` -> ``STOPPED`` transition.
        """
        self.reset_button.setEnabled(True)
        self.start_button.set_icon(ControlBarIcon.START)
        self.stop_button.setEnabled(False)

    def pause(self) -> None:
        """Method called in the ``RUNNING`` -> ``PAUSED`` transition.
        """
        self.start_button.set_icon(ControlBarIcon.START)

    def resume(self) -> None:
        """Method called in the ``PAUSED -> ``RUNNING`` transition.
        """
        self.start_button.set_icon(ControlBarIcon.PAUSE)

    def stop(self) -> None:
        """Method called in the ``PAUSED`` -> ``STOPPED`` transition.
        """
        self.reset_button.setEnabled(True)
        self.start_button.set_icon(ControlBarIcon.START)
        self.stop_button.setEnabled(False)

    def set_reset(self) -> None:
        """Overloaded method.
        """
        FiniteStateMachineLogic.set_reset(self)
        self.set_reset_triggered.emit()

    def set_stopped(self) -> None:
        """Overloaded method.
        """
        FiniteStateMachineLogic.set_stopped(self)
        self.set_stopped_triggered.emit()

    def set_running(self) -> None:
        """Overloaded method.
        """
        FiniteStateMachineLogic.set_running(self)
        self.set_running_triggered.emit()

    def set_paused(self) -> None:
        """Overloaded method.
        """
        FiniteStateMachineLogic.set_paused(self)
        self.set_paused_triggered.emit()


class SimpleControlBar(ControlBar):

    """Simpler version of a control bar, where essentially we prevent the possibility
    to press the pause button.

    This is a small tweak to go along with simple data acquisition systems (such
    as plasduino) where on the hardware side there is essentially no good way to
    pause the system---all we can do is to start and stop the data acquisition.

    .. warning::
        This has not been thoroughly tested, and this derived class is still
        inheriting all the complex logic of its base class, but I do think that
        just re-implementing these two methods will do the trick.
    """

    def start_run(self) -> None:
        """Method called in the ``STOPPED`` -> ``RUNNING`` transition.
        """
        self.reset_button.setEnabled(False)
        self.start_button.setEnabled(False)
        self.stop_button.setEnabled(True)

    def stop_run(self) -> None:
        """Method called in the ``RUNNING`` -> ``STOPPED`` transition.
        """
        self.reset_button.setEnabled(True)
        self.start_button.setEnabled(True)
        self.stop_button.setEnabled(False)


class MainWindow(QtWidgets.QMainWindow):

    """Base class for a DAQ main window.
    """

    _PROJECT_NAME = None
    _MINIMUM_WIDTH = 1000
    _TAB_ICON_SIZE = QtCore.QSize(25, 25)
    _CONTROL_BAR_CLASS = ControlBar

    def __init__(self, parent: QtWidgets.QWidget = None) -> None:
        """Constructor.
        """
        if self._PROJECT_NAME is None:
            msg = f"{self.__class__.__name__} needs to be subclassed, and _PROJECT_NAME set."
            raise RuntimeError(msg)
        super().__init__(parent)
        self.setCentralWidget(QtWidgets.QWidget())
        self.centralWidget().setLayout(QtWidgets.QGridLayout())
        self.centralWidget().setMinimumWidth(self._MINIMUM_WIDTH)
        self.control_bar = self._CONTROL_BAR_CLASS(self)
        self.add_widget(self.control_bar, 1, 0)
        self.run_control_card = RunControlCard()
        self.run_control_card.set(RunControlCardField.PROJECT_NAME, self._PROJECT_NAME)
        self.add_widget(self.run_control_card, 0, 0)
        self.tab_widget = QtWidgets.QTabWidget()
        # self.tab_widget.setTabPosition(self.tab_widget.TabPosition.West)
        self.tab_widget.setIconSize(self._TAB_ICON_SIZE)
        self.add_widget(self.tab_widget, 0, 1, 2, 1)
        self.event_handler_card = EventHandlerCard()
        self.add_tab(self.event_handler_card, "Event handler")
        self.settings_widget = SettingsConfigurationWidget()
        self.add_tab(self.settings_widget, "Settings")
        self.user_application_widget = ConfigurationSectionWidget()
        self.add_tab(self.user_application_widget, "User application")
        self.run_control = None
        self._config_class = None

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        """Overloaded method to avoid closing the main GUI with the run control
        still running, see https://github.com/lucabaldini/baldaquin/issues/28.
        """
        # pylint: disable = invalid-name
        if self.run_control is not None:
            self.run_control.force_reset()
            event.accept()

    @staticmethod
    def open_url(url: str) -> None:
        """Open a local URL using the underlying OS facilities.

        See https://doc.qt.io/qtforpython-6/PySide6/QtGui/QDesktopServices.html
        for more details.
        """
        QtGui.QDesktopServices.openUrl(url)

    def add_widget(self, widget: QtWidgets.QWidget, row: int, col: int,
                   row_span: int = 1, col_span: int = 1) -> None:
        """Add a widget to the underlying layout.

        This is just a convenience method mimicking the corresponding hook of the
        underlying layout.

        Arguments
        ---------
        widget : QtWidgets.QWidget
            The widget to be added to the underlying layout.

        row : int
            The starting row position for the widget.

        col : int
            The starting column position for the widget.

        row_span : int, optional (default 1)
            The number of rows spanned by the widget.

        col_span : int, optional (default 1)
            The number of columns spanned by the widget.
        """
        # pylint: disable=too-many-arguments
        self.centralWidget().layout().addWidget(widget, row, col, row_span, col_span)

    def add_tab(self, page: QtWidgets.QWidget, label: str,
                icon_name: str = None) -> QtWidgets.QWidget:
        """Add a page to the tab widget.

        Arguments
        ---------
        page : QtWidgets.QWidget
            The widget to be added to the tab widget.

        label : str
            The text label to be displayed on the tab.

        icon_name : str, optional
            The name of the icon to be displayed on the tab (if None, non icon is shown).
        """
        pos = self.tab_widget.addTab(page, label)
        if icon_name is not None:
            self.tab_widget.setTabIcon(pos, load_icon(icon_name))
        return page

    def add_plot_canvas_tab(self, label: str, update_interval: int = 500,
                            icon_name: str = None, **kwargs):
        """Add a page to the tab widget holding a matplotlib plot.

        Note that the proper signals of the control bar are automatically connected
        to the start_updating and stop_updating slots of the PlotCanvasWidget object
        that is being added to the tab.

        Arguments
        ---------
        label : str
            The text label to be displayed on the tab.

        update_interval : int
            The update interval in ms.

        icon_name : str, optional
            The name of the icon to be displayed on the tab (if None, non icon is shown).

        kwargs : dict
            All the keyword arguments to be passed to the matplotlib subplots() call.
        """
        widget = PlotCanvasWidget(update_interval, **kwargs)
        self.control_bar.set_running_triggered.connect(widget.start_updating)
        self.control_bar.set_stopped_triggered.connect(widget.stop_updating)
        return self.add_tab(widget, label, icon_name)

    def update_run_control_test_stand_id(self, test_stand_id: int) -> None:
        """Update the test stand ID in the run control card.
        """
        self.run_control_card.set(RunControlCardField.TEST_STAND_ID, test_stand_id)

    def update_run_control_run_id(self, run_id: int) -> None:
        """Update the test run ID in the run control card.
        """
        self.run_control_card.set(RunControlCardField.RUN_ID, run_id)

    def update_run_control_state(self, state: FsmState) -> None:
        """Update the test run control state in the run control card.
        """
        self.run_control_card.set(RunControlCardField.STATE, state.value)

    def update_run_control_uptime(self, value: float) -> None:
        """Update the uptime in the run contro card.
        """
        self.run_control_card.set(RunControlCardField.UPTIME, value)

    def update_event_handler_output_file(self, file_path: Path):
        """Update the output file path in the event handler card.
        """
        self.event_handler_card.set_file_path(file_path)

    def update_event_handler_stats(self, statistics: PacketStatistics, event_rate: float) -> None:
        """Update the data taking statistics in the event handler card.
        """
        self.event_handler_card.set(EventHandlerCardField.NUM_PACKETS_PROCESSED,
                                    statistics.packets_processed)
        self.event_handler_card.set(EventHandlerCardField.NUM_PACKETS_WRITTEN,
                                    statistics.packets_written)
        self.event_handler_card.set(EventHandlerCardField.NUM_BYTES_WRITTEN,
                                    statistics.bytes_written)
        self.event_handler_card.set(EventHandlerCardField.AVERAGE_EVENT_RATE, event_rate)

    def set_run_control(self, run_control: RunControlBase) -> None:
        """Set the child run control that the GUI should manage.
        """
        self.run_control = run_control
        # Run a first update of the GUI elements that need to be set...
        self.update_run_control_test_stand_id(self.run_control.test_stand_id())
        self.update_run_control_run_id(self.run_control.run_id())
        self.update_run_control_state(self.run_control.state())
        # ... and setup the connections that guarantee that these elemnts remain up to date.
        self.run_control.run_id_changed.connect(self.update_run_control_run_id)
        self.run_control.state_changed.connect(self.update_run_control_state)
        self.run_control.uptime_updated.connect(self.update_run_control_uptime)
        self.run_control.event_handler_stats_updated.connect(self.update_event_handler_stats)
        # Fully connect the control bar to the run control. Note that, while for
        # most of the transitions we can map the control bar signals and the
        # run control slots directly, we need a custom slot for the start run
        # because, depending on whether we go through the start_run() or the resume()
        # FSM hooks, we need to apply the configuration that is stored in the GUI.
        self.control_bar.set_reset_triggered.connect(self.run_control.set_reset)
        self.control_bar.set_stopped_triggered.connect(self.run_control.set_stopped)
        self.control_bar.set_running_triggered.connect(self.set_run_control_running)
        self.control_bar.set_paused_triggered.connect(self.run_control.set_paused)
        # Be prepared for when a user application is loaded on the run control.
        self.run_control.user_application_loaded.connect(self.setup_user_application)

    def setup_user_application(self, user_application: UserApplicationBase) -> None:
        """Setup the user application.
        """
        # Update the proper (static) GUI elements.
        self.run_control_card.set(RunControlCardField.USER_APPLICATION, user_application.NAME)
        buffer_class = user_application.event_handler.BUFFER_CLASS.__name__
        self.event_handler_card.set(EventHandlerCardField.BUFFER_CLASS, buffer_class)
        # Display the application configuration.
        # Keep a reference to the input configuration class so that we can return
        # the current (possibly modified) configuration as an instance of the
        # proper class.
        self._config_class = user_application.configuration.__class__
        self.settings_widget.display(user_application.configuration)
        self.user_application_widget.display(user_application.configuration.application_section())
        # Connect the necessary signal for keeping the thing in synch.
        user_application.event_handler.output_file_set.connect(
            self.update_event_handler_output_file)
        # Setup the control bar in the STOPPED state---note that we are not calling
        # the set_stopped() hook, here, as we don't want to trigger another set_stopped()
        # action on the run control side.
        self.control_bar.set_state(FsmState.STOPPED)
        self.control_bar.setup()

    def current_configuration(self) -> UserApplicationConfiguration:
        """Return the current user application configuration.
        """
        config = self._config_class()
        for section in (self.settings_widget.logging_widget.current_configuration_section(),
                        self.settings_widget.buffering_widget.current_configuration_section(),
                        self.settings_widget.multicast_widget.current_configuration_section(),
                        self.user_application_widget.current_configuration_section()):
            config.overwrite_section(section)
        return config

    def set_run_control_running(self) -> None:
        """Custom slot to set the run control in the RUNNING state.

        The important bit, here, is that, when we start the run control from the
        STOPPED state, we apply the user application configuration currently displayed
        in the GUI.
        """
        if self.run_control.is_stopped():
            configuration = self.current_configuration()
            self.run_control.configure_user_application(configuration)
        self.run_control.set_running()


def bootstrap_qapplication():
    """Create a QApplication object and apply the proper stylesheet.
    """
    # pylint: disable=unspecified-encoding
    qapp = QtWidgets.QApplication(sys.argv)
    with open(stylesheet_file_path()) as stylesheet:
        qapp.setStyleSheet(stylesheet.read())
    return qapp


def bootstrap_window(window_class, run_control: RunControlBase = None,
                     user_application: UserApplicationBase = None) -> None:
    """Bootstrap a main window.

    This is creating a QApplication, applying the relevant stylesheet, and
    creating an actual instance of the window class passed as an argument.
    """
    qapp = bootstrap_qapplication()
    window = window_class()
    if run_control is not None:
        window.set_run_control(run_control)
        if user_application is not None:
            run_control.load_user_application(user_application)
    window.show()
    exec_qapp(qapp)



# Copyright (C) 2024--2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Basic definition of the plasduino communication protocol.
"""

from enum import IntEnum

from baldaquin.pkt import AbstractPacket, FixedSizePacketBase, Format, Layout, packetclass

COMMENT_PREFIX = '# '
TEXT_SEPARATOR = ', '


class Marker(IntEnum):

    """Relevant protocol markers, verbatim from
    https://bitbucket.org/lbaldini/plasduino/src/master/arduino/protocol.h

    (In the old days we used to have this generated automatically from the
    corresponding header file, but the project is so stable now, that this seems
    hardly relevant.)
    """

    NO_OP_HEADER = 0xa0
    DIGITAL_TRANSITION_HEADER = 0xa1
    ANALOG_READOUT_HEADER = 0xa2
    GPS_MEASSGE_HEADER = 0xa3
    RUN_END_MARKER = 0xb0


class OpCode(IntEnum):

    """Definition of the operational codes, verbatim from
    https://bitbucket.org/lbaldini/plasduino/src/master/arduino/protocol.h

    (In the old days we used to have this generated automatically from the
    corresponding header file, but the project is so stable now, that this seems
    hardly relevant.)
    """

    OP_CODE_NO_OP = 0x00
    OP_CODE_START_RUN = 0x01
    OP_CODE_STOP_RUN = 0x02
    OP_CODE_SELECT_NUM_DIGITAL_PINS = 0x03
    OP_CODE_SELECT_DIGITAL_PIN = 0x04
    OP_CODE_SELECT_NUM_ANALOG_PINS = 0x05
    OP_CODE_SELECT_ANALOG_PIN = 0x06
    OP_CODE_SELECT_SAMPLING_INTERVAL = 0x07
    OP_CODE_SELECT_INTERRUPT_MODE = 0x08
    OP_CODE_SELECT_PWM_DUTY_CYCLE = 0x09
    OP_CODE_SELECT_POLLING_MODE = 0x0a
    OP_CODE_AD9833_CMD = 0x0b
    OP_CODE_TOGGLE_LED = 0x0c
    OP_CODE_TOGGLE_DIGITAL_PIN = 0x0d


class InterruptMode(IntEnum):

    """Definition of the interrupt modes.
    """

    DISABLED = 0
    CHANGE = 1
    FALLING = 2
    RISING = 3


@packetclass
class PlasduinoPacketBase(FixedSizePacketBase):

    """Base class for the plasduino packets.
    """

    OUTPUT_HEADERS = None
    OUTPUT_ATTRIBUTES = None
    OUTPUT_FMTS = None

    @classmethod
    def text_header(cls, prefix: str = COMMENT_PREFIX, creator: str = None) -> str:
        """Return the header for the output text file.
        """
        headers = ', '.join(map(str, cls.OUTPUT_HEADERS))
        return f'{AbstractPacket.text_header(prefix, creator)}{prefix}{headers}\n'

    def to_text(self, separator: str = TEXT_SEPARATOR) -> str:
        """Convert a readout to text for use in a custom sink.
        """
        return self._text(self.OUTPUT_ATTRIBUTES, self.OUTPUT_FMTS, separator)

    def __str__(self):
        """String formatting.
        """
        return self._repr(self.OUTPUT_ATTRIBUTES, self.OUTPUT_FMTS)


@packetclass
class AnalogReadout(PlasduinoPacketBase):

    """A plasduino analog readout is a 8-byte binary array containing:

    * byte(s) 0  : the array header (``Marker.ANALOG_READOUT_HEADER``);
    * byte(s) 1  : the analog pin number;
    * byte(s) 2-5: the timestamp of the readout from millis();
    * byte(s) 6-7: the actual adc value.
    """

    layout = Layout.BIG_ENDIAN
    header: Format.UNSIGNED_CHAR = Marker.ANALOG_READOUT_HEADER
    pin_number: Format.UNSIGNED_CHAR
    milliseconds: Format.UNSIGNED_LONG
    adc_value: Format.UNSIGNED_SHORT

    OUTPUT_HEADERS = ('Pin number', 'Time [s]', 'ADC counts')
    OUTPUT_ATTRIBUTES = ('pin_number', 'seconds', 'adc_value')
    OUTPUT_FMTS = ('%d', '%.3f', '%d')

    def __post_init__(self) -> None:
        """Post initialization.
        """
        self.seconds = 1.e-3 * self.milliseconds


@packetclass
class DigitalTransition(PlasduinoPacketBase):

    """A plasduino digital transition is a 6-byte binary array containing:

    * byte(s) 0  : the array header (``Marker.DIGITAL_TRANSITION_HEADER``);
    * byte(s) 1  : the transition information (pin number and edge type);
    * byte(s) 2-5: the timestamp of the readout from micros().
    """

    layout = Layout.BIG_ENDIAN
    header: Format.UNSIGNED_CHAR = Marker.DIGITAL_TRANSITION_HEADER
    info: Format.UNSIGNED_CHAR
    microseconds: Format.UNSIGNED_LONG

    OUTPUT_HEADERS = ('Pin number', 'Edge type', 'Time [s]')
    OUTPUT_ATTRIBUTES = ('pin_number', 'edge', 'seconds')
    OUTPUT_FMTS = ('%d', '%d', '%.6f')

    def __post_init__(self) -> None:
        """Post initialization.
        """
        # Note the _info field is packing into a single byte the edge type
        # (the MSB) and the pin number.
        self.pin_number = self.info & 0x7F
        self.edge = (self.info >> 7) & 0x1
        self.seconds = 1.e-6 * self.microseconds



# Copyright (C) 2024 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Plasduino common resources.
"""

from __future__ import annotations

import struct
import time
from typing import Any

from aptapy.strip import StripChart

from baldaquin import arduino_, plasduino
from baldaquin.app import UserApplicationBase
from baldaquin.config import UserApplicationConfiguration
from baldaquin.event import EventHandlerBase
from baldaquin.logging_ import logger
from baldaquin.plasduino.protocol import (
    AnalogReadout,
    DigitalTransition,
    InterruptMode,
    Marker,
    OpCode,
)
from baldaquin.plasduino.sketches import sketch_file_path
from baldaquin.runctrl import RunControlBase
from baldaquin.serial_ import SerialInterface

# List of supported boards, i.e., only the arduino uno at the moment.
_SUPPORTED_BOARDS = (arduino_.UNO, )


class PlasduinoSerialInterface(SerialInterface):

    """Specialized plasduino serial interface.

    This is derived class of our basic serial interface, where we essentially
    implement the simple plasduino communication protocol.
    """

    # pylint: disable=too-many-ancestors

    def read_and_unpack(self, fmt: str) -> Any:
        """Overloaded function.

        For some reason on the arduino side we go into the trouble of reverting the
        native bit order and we transmit things as big endian---I have no idea
        what I was thinking, back in the days, but I don't think it makes sense
        to fix this nonsense, now. Extra work on the transmitting side, and extra
        work on the receiving side too. Good job, Luca!
        """
        return super().read_and_unpack(f">{fmt}")

    def read_sketch_info(self):
        """Read the information about the sketch loaded on the board.

        The sketch information (identifier, version) is the first bit of info that
        the plasduino sketches send out, and this should be the first function
        called after a setup.
        """
        sketch_id = self.read_and_unpack("B")
        sketch_version = self.read_and_unpack("B")
        return sketch_id, sketch_version

    def read_run_end_marker(self) -> None:
        """Read a single byte from the serial port and make sure it is the
        run-end marker.

        (The run end marker is sent over through the serial port from the arduino
        sketch when the run is stopped and all the measurements have been completed,
        and needs to be taken out of the way in order for the next run to take place.)
        """
        logger.info("Waiting for the run end marker...")
        marker = self.read_and_unpack("B")
        if not marker == Marker.RUN_END_MARKER:
            raise RuntimeError(f"Run end marker mismatch "
                  f"(expected {hex(Marker.RUN_END_MARKER)}, found {hex(marker)}).")
        logger.info("Run end marker correctly read.")

    def read_until_run_end_marker(self, timeout: float = None) -> None:
        """Read data from the serial port until the end-of-run marker is found.

        This is actually never used, as the intermediate data would get lost, but
        it is potentially useful when debugging the serial communication.

        Arguments
        ---------
        timeout : float (default None)
            The timeout (in s) to be temporarily set for the transaction.
        """
        logger.info("Scanning serial input for run-end marker...")
        previous_timeout = self.timeout
        if timeout != self.timeout:
            self.timeout = timeout
            logger.debug(f"Serial port timeout temporarily set to {self.timeout} s...")
        data = self.read_until(struct.pack("B", Marker.RUN_END_MARKER))
        if len(data) > 0:
            logger.debug(f"{len(data)} byte(s) found: {data}")
        if previous_timeout != self.timeout:
            self.timeout = previous_timeout
            logger.debug(f"Serial port timeout restored to {self.timeout} s...")

    def write_opcode(self, opcode: OpCode) -> int:
        """Write the value of a given opcode to the serial port.

        This is typically meant to signal the start/stop run, or to configure the
        behavior of the sketch on the arduino side (e.g., select the pins for
        analog readout).

        Arguments
        ---------
        opcode : OpCode
            The operational code to be written to the serial port.
        """
        logger.debug(f"Writing {opcode} to the serial port...")
        return self.pack_and_write(opcode.value, "B")

    def write_start_run(self) -> int:
        """ Write a start run command to the serial port.
        """
        return self.write_opcode(OpCode.OP_CODE_START_RUN)

    def write_stop_run(self) -> int:
        """ Write a stop run command to the serial port.
        """
        return self.write_opcode(OpCode.OP_CODE_STOP_RUN)

    def write_cmd(self, opcode: OpCode, value: int, fmt: str) -> None:
        """ Write a command to the arduino board.

        This implies writing the opcode to the serial port, writing the actual
        payload and, finally, reading back the arduino response and making
        sure the communication went fine.

        And, looking back at this after many years, I cannot help noticing that
        it looks a little bit funny, but I guess it did make sense, back in the
        days.

        Arguments
        ---------
        opcode : OpCode
            The opcode defining the command.

        value : int
            The actual value.

        fmt : str
            The format string.
        """
        self.write_opcode(opcode)
        logger.debug(f"Writing configuration value {value} to the serial port")
        self.pack_and_write(value, fmt)
        target_opcode = self.read_and_unpack("B")
        actual_opcode = self.read_and_unpack("B")
        actual_value = self.read_and_unpack(fmt)
        logger.debug(f"Board response ({target_opcode}, {actual_opcode}, {actual_value})...")
        if actual_opcode != opcode or actual_value != value:
            raise RuntimeError(f"Write/read mismatch in {self.__class__.__name__}.write_cmd()")

    def setup_analog_sampling_sketch(self, pins: list[int], sampling_interval: int) -> None:
        """Setup the `analog_sampling` sketch.

        Note that we are taking a minimal approach, here, where exactly two input
        analog pins are used, and they are those dictated by the Lab 1 shield, i.e.,
        the only thing that we are setting, effectively, is the sampling interval.

        Arguments
        ---------
        sampling_interval : int
            The sampling interval in ms.
        """
        self.write_cmd(OpCode.OP_CODE_SELECT_NUM_ANALOG_PINS, len(pins), "B")
        for pin in pins:
            self.write_cmd(OpCode.OP_CODE_SELECT_ANALOG_PIN, pin, "B")
        self.write_cmd(OpCode.OP_CODE_SELECT_SAMPLING_INTERVAL, sampling_interval, "I")

    def setup_digital_timer_sketch(self, interrupt_mode0, interrupt_mode1) -> None:
        """Setup the `digital_timer` sketch.
        """
        self.write_cmd(OpCode.OP_CODE_SELECT_INTERRUPT_MODE, interrupt_mode0, "B")
        self.write_cmd(OpCode.OP_CODE_SELECT_INTERRUPT_MODE, interrupt_mode1, "B")


class PlasduinoRunControl(RunControlBase):

    """Specialized plasduino run control.
    """

    _PROJECT_NAME = plasduino.PROJECT_NAME


class PlasduinoEventHandlerBase(EventHandlerBase):

    """Plasduino basic event handler.

    This takes care of all the operations connected with the handshaking and
    sketch upload. Derived classes must implement the ``read_packet()`` slot.
    """

    # pylint: disable=abstract-method
    SKETCH_ID = None
    SKETCH_VERSION = None

    def __init__(self) -> None:
        """Constructor.

        Note we create an empty serial interface, here, and we then open the port
        while setting up the user application.
        """
        super().__init__()
        self.serial_interface = PlasduinoSerialInterface()

    def open_serial_interface(self, timeout: float = None, handshake_timeout: float = 5.) -> None:
        """Autodetect a supported arduino board, open the serial connection to it,
        and do the handshaking.

        Arguments
        ---------
        timeout : float (default None)
            The timeout (in s) for the serial readout. If set to None, every read
            operation is effectively blocking, and this is the way we should operate
            in normal conditions.

        handshake_timeout : float
            The timeout for the first handshaking with the board, when we try and
            gauge the identifier and version of the sketch that is preloaded.
            Note that this cannot be too small, as it takes a good second for the
            thing to happen.
        """
        port_info = arduino_.autodetect_arduino_board(*_SUPPORTED_BOARDS)
        if port_info is None:
            raise RuntimeError("Could not find a suitable arduino board connected.")
        self.serial_interface.connect(port_info, timeout=timeout)
        self.serial_interface.pulse_dtr()
        logger.info("Hand-shaking with the arduino board...")

        # Temporarily set a finite timeout to handle the case where there is not
        # sensible sketch pre-loaded on the board, and we have to start from scratch.
        self.serial_interface.timeout = handshake_timeout
        try:
            sketch_id, sketch_version = self.serial_interface.read_sketch_info()
            logger.info(f"Sketch {sketch_id} version {sketch_version} loaded onboard...")
        except struct.error:
            logger.warning("There seems to be no plasduino scketch pre-loaded on the board...")
            sketch_id, sketch_version = None, None
        # Now put back the actual target timeout.
        self.serial_interface.timeout = timeout

        # If the sketch uploaded onboard is the one we expect, we're good to go.
        if (sketch_id, sketch_version) == (self.SKETCH_ID, self.SKETCH_VERSION):
            return

        # Otherwise we have to upload the proper sketch.
        file_path = sketch_file_path(self.SKETCH_ID, self.SKETCH_VERSION)
        board = arduino_.ArduinoBoard.by_device_id(port_info.device_id)
        arduino_.upload_sketch(file_path, board.designator, port_info.name)
        sketch_id, sketch_version = self.serial_interface.read_sketch_info()
        if (sketch_id, sketch_version) != (self.SKETCH_ID, self.SKETCH_VERSION):
            raise RuntimeError(f"Could not upload sketch {self.SKETCH_ID} "
                               f"version {self.SKETCH_VERSION}")

    def close_serial_interface(self) -> None:
        """Close the serial interface.
        """
        self.serial_interface.disconnect()


class PlasduinoAnalogEventHandler(PlasduinoEventHandlerBase):

    """Event handler for the plasduino sketches reading analog data.
    """

    SKETCH_ID = 2
    SKETCH_VERSION = 3

    def read_packet(self) -> int:
        """Read a single packet, that is, an analog readout.
        """
        return self.serial_interface.read(AnalogReadout.size)

    def wait_pending_packets(self, wait_time: int = None) -> int:
        """Wait and read all the pending packets from the serial port, then consume
        the run end marker.

        This is necessary because, after the sketch running on the arduino board
        receives the run end opcode, there is a varible number of analog readouts
        (ranging from 0 to 2 if 2 pins are used) that are acquired before the
        data acquisition is actually stopped. (Yes, poor design on the sketch side.)
        What we do here is essentially: wait a long enough time (it should be
        equal or longer than the sampling time to catch all the corner cases),
        see how many bytes are waiting in the input buffer of the serial port,
        calculate the number of pending packets, read them and finally consume the
        run end marker, so that we are ready to start again.

        Note that the pending packets are correctly processed, passed to the
        event handler buffer and written to disk.

        Arguments
        ---------
        wait_time : int (default None)
            The amount of time (in ms) we wait before polling the serial port
            for additional pending packets.
        """
        logger.info(f"Waiting {wait_time} ms for pending packet(s)...")
        if wait_time is not None:
            time.sleep(wait_time / 1000.)
        num_bytes = self.serial_interface.in_waiting
        # At this point we expect a number of events which is a multiple of
        # AnalogReadout.size, + 1. If this is not the case, it might indicate that
        # we have not waited enough.
        if num_bytes % AnalogReadout.size != 1:
            logger.warning(f"{num_bytes} pending bytes on the serial port, expected 1, 9 or 17...")
        num_packets = num_bytes // AnalogReadout.size
        if num_packets > 0:
            logger.info(f"Reading the last {num_packets} packet(s) from the serial port...")
            for _ in range(num_packets):
                self.acquire_packet()
            self.flush_buffer()
        self.serial_interface.read_run_end_marker()


class PlasduinoDigitalEventHandler(PlasduinoEventHandlerBase):

    """Event handler for the plasduino sketches reading digital data.
    """

    SKETCH_ID = 1
    SKETCH_VERSION = 3

    def read_packet(self):
        """Read a single packet, that is, an analog readout.
        """
        return self.serial_interface.read(DigitalTransition.size)


class PlasduinoAnalogConfiguration(UserApplicationConfiguration):

    """User application configuration for plasduino analog applications.
    """

    _PARAMETER_SPECS = (
        ("strip_chart_max_length", int, 200, "Strip chart maximum length",
            dict(min=10, max=1000000)),
    )


class PlasduinoDigitalConfiguration(UserApplicationConfiguration):

    """User application configuration for plasduino digital applications.
    """

    _PARAMETER_SPECS = ()


class PlasduinoUserApplicationBase(UserApplicationBase):

    """Base class for all the plasduino applications.
    """

    def configure(self):
        """Overloaded method.
        """

    def teardown(self) -> None:
        """Overloaded method (STOPPED -> RESET).
        """
        self.event_handler.close_serial_interface()

    def start_run(self) -> None:
        """Overloaded method (STOPPED -> RUNNING).
        """
        self.event_handler.serial_interface.write_start_run()
        super().start_run()


class PlasduinoAnalogUserApplicationBase(PlasduinoUserApplicationBase):

    """Specialized base class for plasduino user applications relying on the
    `analog_sampling` sketch.
    """

    _PINS = None
    _SAMPLING_INTERVAL = None
    _ADDITIONAL_PENDING_WAIT = 200

    @staticmethod
    def create_strip_charts(pins: list[int], ylabel: str = "ADC counts"):
        """Create all the strip charts for displaying real-time data.
        """
        kwargs = dict(xlabel="Time [s]", ylabel=ylabel)
        return {pin: StripChart(label=f"Pin {pin}", **kwargs) for pin in pins}

    def setup(self) -> None:
        """Overloaded method (RESET -> STOPPED).
        """
        self.event_handler.open_serial_interface()
        self.event_handler.serial_interface.setup_analog_sampling_sketch(self._PINS,
                                                                         self._SAMPLING_INTERVAL)

    def stop_run(self) -> None:
        """Overloaded method (RUNNING -> STOPPED).
        """
        super().stop_run()
        self.event_handler.serial_interface.write_stop_run()
        self.event_handler.wait_pending_packets(self._SAMPLING_INTERVAL +
                                                self._ADDITIONAL_PENDING_WAIT)


class PlasduinoDigitalUserApplicationBase(PlasduinoUserApplicationBase):

    """Specialized base class for plasduino user applications relying on the
    `digital_timer` sketch.
    """

    def setup(self) -> None:
        """Overloaded method (RESET -> STOPPED).
        """
        self.event_handler.open_serial_interface()
        args = InterruptMode.CHANGE, InterruptMode.DISABLED
        self.event_handler.serial_interface.setup_digital_timer_sketch(*args)

    def stop_run(self) -> None:
        """Overloaded method (RUNNING -> STOPPED).
        """
        super().stop_run()
        self.event_handler.serial_interface.write_stop_run()
        self.event_handler.serial_interface.read_run_end_marker()



# Copyright (C) 2022 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The glorious baldaquin logo :-)
"""

import os

import matplotlib.pyplot as plt
import numpy as np
from scipy import interpolate

from baldaquin import BALDAQUIN_DOCS_STATIC

FIG_SIDE = 6
MODES = ('light', 'dark')
FILL_COLOR = 'skyblue'


def _line_color(mode):
    """Get the line color for the theme.
    """
    assert mode in MODES
    if mode == 'light':
        return 'lightgray'
    if mode == 'dark':
        return 'white'


def plot_spline(*nodes, control_points=False, **kwargs):
    """Generic function to plot a spline.
    """
    xp = np.array(nodes)[:, 0]
    yp = np.array(nodes)[:, 1]
    tck, u = interpolate.splprep([xp, yp], s=0)
    x, y = interpolate.splev(np.linspace(0., 1., 100), tck, der=0)
    plt.plot(x, y, **kwargs)
    if control_points:
        plt.plot(xp, yp, 'o')
    return x, y


def figure(x1=-1.1, x2=1.1, y1=-1.1, y2=1.1):
    """Crete the logo figure.
    """
    plt.figure(figsize=(FIG_SIDE, FIG_SIDE))
    plt.gca().set_aspect('equal')
    plt.gca().axis((x1, x2, y1, y2))
    plt.subplots_adjust(left=0., right=1., bottom=0., top=1.)
    plt.axis('off')


def plot_baldaquin(line_color, lw, w=0.425, h=0.40, wave=True):
    """Plot the glorious baldaquin.
    """
    kwargs = dict(color=line_color, lw=lw)
    plt.plot((-w, -w, w, w), (-h, h, h, -h), **kwargs)
    x, y = plot_spline((0., 0.99 * h), (-0.05 * w, 0.8 * h),
                       (-0.4 * w, 0.5 * h), (-w, 0.), **kwargs)
    plt.fill_between(x, y, np.full(x.shape, h), color=FILL_COLOR)
    x, y = plot_spline((0., 0.99 * h), (0.05 * w, 0.8 * h), (0.4 * w, 0.5 * h), (w, 0.), **kwargs)
    plt.fill_between(x, y, np.full(x.shape, h), color=FILL_COLOR)
    plot_spline((-1.25 * w, 0.9 * h), (-w, h), (-0.5 * w, 1.25 * h), (0., 1.75 * h), **kwargs)
    plot_spline((1.25 * w, 0.9 * h), (w, h), (0.5 * w, 1.25 * h), (0., 1.75 * h), **kwargs)
    if wave:
        x = np.linspace(-0.8 * w, 0.8 * w, 100)
        y = -1.25 * h + 0.25 * h * np.cos(30 * x) * np.exp(-3. * x)
        plt.plot(x, y, color=line_color, lw=lw)


def create_logo(mode, lw=3.5, circle_radius=1.):
    """Create the big logo.
    """
    line_color = _line_color(mode)
    figure()
    circle = plt.Circle((0, 0), circle_radius, color=line_color, lw=3.5, fill=False)
    plt.gca().add_patch(circle)
    plt.text(0., 0., 'bal  daq  uin', ha='center', va='center', size=60, color=FILL_COLOR)
    plot_baldaquin(line_color, lw)
    text = 'BALd DAQ User INterface'
    max_angle = np.radians(55.)
    num_letters = len(text)
    radius = 0.925 * circle_radius
    for i, letter in enumerate(text):
        angle = -max_angle + 2. * max_angle * (i / (num_letters - 1)) - 0.5 * np.pi
        x, y = radius * np.cos(angle), radius * np.sin(angle)
        plt.text(x, y, letter, size=20., rotation=np.degrees(angle) + 90., ha='center',
                 va='center', color=FILL_COLOR)
    file_name = f'baldaquin_logo_{mode}.png'
    plt.savefig(os.path.join(BALDAQUIN_DOCS_STATIC, file_name), transparent=True)


def create_small_logo(mode, lw=15.):
    """Create the small logo.
    """
    line_color = _line_color(mode)
    figure(-0.54, 0.54, -0.36, 0.72)
    plot_baldaquin(line_color=line_color, lw=lw, wave=False)
    file_name = f'baldaquin_logo_small_{mode}.png'
    plt.savefig(os.path.join(BALDAQUIN_DOCS_STATIC, file_name), transparent=True)


for mode in MODES:
    create_logo(mode)
    create_small_logo(mode)


plt.show()



# Copyright (C) 2024 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for egu.py
"""

import numpy as np

from baldaquin import egu
from baldaquin.plasduino import PLASDUINO_SENSORS


def test_linear(slope: float = 2., intercept: float = 1.):
    """Test a simple linear conversion.
    """
    converter = egu.LinearConversion(slope, intercept)
    raw = np.linspace(0., 1., 11)
    physical = converter(raw)
    assert np.allclose(physical, slope * raw + intercept)


def test_thermistor():
    """Test a spline conversion.
    """
    file_path = PLASDUINO_SENSORS / "NXFT15XH103FA2B.dat"
    _ = egu.ThermistorConversion.from_file(file_path, 10., 10, col_resistance=2)



# Copyright (C) 2022 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for buf.py
"""

import io
import os
from pathlib import Path

import pytest

from baldaquin.buf import FIFO, CircularBuffer, Sink, WriteMode
from baldaquin.env import BALDAQUIN_DATA, BALDAQUIN_ENCODING
from baldaquin.pkt import FixedSizePacketBase, Format, Layout, packetclass


@packetclass
class Packet(FixedSizePacketBase):

    """Plausible data structure for testing purposes.
    """

    layout = Layout.BIG_ENDIAN
    header: Format.UNSIGNED_CHAR = 0xaa
    packet_id: Format.UNSIGNED_SHORT

    def as_text(self):
        """Simple text formatting function.
        """
        return f"{self.header}, {self.packet_id}\n"


def _test_buffer_base(buffer_class, num_packets: int = 10, **kwargs):
    """Base function to test a generic, concrete subclass
    """
    buffer = buffer_class(**kwargs)
    for i in range(num_packets):
        packet = Packet(Packet.header, i)
        buffer.put(packet)
    assert buffer.size() == num_packets
    # Since we have no sink connected, any flush attempt should raise an exception.
    with pytest.raises(RuntimeError):
        buffer.flush()
    buffer.clear()


def test_fifo():
    """Test a FIFO
    """
    _test_buffer_base(FIFO)


def test_circular_buffer():
    """Test a circular buffer.
    """
    _test_buffer_base(CircularBuffer)


def _scrap_file(file_path: Path) -> Path:
    """Make sure that a given file does not exist, and remove if it does.
    """
    if file_path.exists():
        os.remove(file_path)
    return file_path


def test_sink_contextmanager():
    """Test the context manager sink protocol.
    """
    file_path = _scrap_file(BALDAQUIN_DATA / "sink.dat")
    # Create the sink.
    sink = Sink(file_path, WriteMode.BINARY)
    # Open the sink.
    with sink.open() as output_file:
        assert isinstance(output_file, io.IOBase)
    # Check that creating another sink with the same file path raises an exception.
    with pytest.raises(FileExistsError):
        sink = Sink(file_path, WriteMode.BINARY)
    # Cleanup.
    os.remove(file_path)


def test_sink_header():
    """Test the header for a text sink.
    """
    file_path = _scrap_file(BALDAQUIN_DATA / "sink.txt")
    header = "#This is a file header.\n"
    # Create the sink---this should write the header into the output file...
    _ = Sink(file_path, WriteMode.TEXT, header=header)
    # ... open the output file and make sure the content is correct.
    with open(file_path, encoding=BALDAQUIN_ENCODING) as input_file:
        assert input_file.read() == header
    # Cleanup.
    os.remove(file_path)


def test_buffer_flush(num_packets: int = 10):
    """Test the full flush mechanism with multiple sinks.
    """
    binary_file_path = _scrap_file(BALDAQUIN_DATA / "sink.dat")
    text_file_path = _scrap_file(BALDAQUIN_DATA / "sink.txt")
    # Create a buffer and assign two sinks.
    buffer = CircularBuffer()
    buffer.set_primary_sink(binary_file_path)
    buffer.add_custom_sink(text_file_path, WriteMode.TEXT, Packet.as_text, "#File header\n")
    # Fill the buffer with some packets.
    for i in range(num_packets):
        packet = Packet(Packet.header, i)
        buffer.put(packet)
    # At this point the buffer should have the target number of packets.
    assert buffer.size() == num_packets
    buffer.flush()
    # And now it should be empty.
    assert buffer.size() == 0
    with open(text_file_path, encoding=BALDAQUIN_ENCODING) as input_file:
        _ = input_file.read()
    # Cleanup
    os.remove(binary_file_path)
    os.remove(text_file_path)



# Copyright (C) 2022 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for timeline.py
"""

import time

from baldaquin.timeline import Timeline, Timestamp


def test_default_timeline():
    """The timestamp seconds of the default timeline should be essentially
    identical to time.time()---mind the two things are necessarily cached at
    different times.
    """
    timeline = Timeline()
    tstamp = timeline.latch()
    delta = tstamp.seconds - time.time()
    assert abs(delta) < 0.001


def test_offset_timeline(origin="1971-01-01"):
    """Test a timeline with a 1-year offset with respect to the POSIX time.
    """
    timeline = Timeline(origin)
    tstamp = timeline.latch()
    delta = tstamp.seconds - time.time()
    assert abs(delta + 3600. * 24 * 365) < 0.001


def test_offset_timeline_leap(origin="1973-01-01"):
    """Test a timeline with a 3-year offset with respect to the POSIX time,
    including one leap year.
    """
    timeline = Timeline(origin)
    tstamp = timeline.latch()
    delta = tstamp.seconds - time.time()
    assert abs(delta + 3600. * 24 * (3 * 365 + 1)) < 0.001


def test_serialization():
    """Test the serialization/deserialization facilities.
    """
    # Create a timeline and a timestamp.
    timeline = Timeline()
    timestamp = timeline.latch()
    # Serialize...
    kwargs = timestamp.to_dict()
    # ... and deserialize.
    twin = Timestamp.from_dict(**kwargs)
    # The two things should be identical.
    assert twin == timestamp
    assert twin.utc_datetime == timestamp.utc_datetime
    assert twin.local_datetime == timestamp.local_datetime
    assert twin.seconds == timestamp.seconds
    assert id(twin) != id(timestamp)
    # And, just for fun, make sure a different timestamp is really different.
    impostor = timeline.latch()
    assert impostor != timestamp
